from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Mapping, Sequence, TYPE_CHECKING, TypeAlias

import numpy as np
import numpy.typing as npt

from .contact_interface import (
    assemble_contact_interface_jacobian as _assemble_contact_interface_jacobian,
    assemble_contact_interface_residual as _assemble_contact_interface_residual,
    assemble_onesided_bilinear,
    assemble_contact_onesided_floor,
    assemble_contact_coupling_matrices as _assemble_contact_coupling_matrices,
    map_surface_facets_to_tet_elements,
    map_surface_facets_to_hex_elements,
)
from .supermesh import build_surface_supermesh
from .surface import SurfaceMesh
from .base import BaseMesh

if TYPE_CHECKING:
    from .contact_interface import ContactCouplingMatrix
    from ..core.weakform import Params as WeakParams
    from .contact_interface import SurfaceMixedFormContext

ContactJacobianReturn: TypeAlias = np.ndarray | tuple[np.ndarray, np.ndarray, np.ndarray, int]
MixedSurfaceResidualForm: TypeAlias = Callable[
    ["SurfaceMixedFormContext", Mapping[str, npt.ArrayLike], Any],
    Mapping[str, npt.ArrayLike],
]
SurfaceHatFn: TypeAlias = Callable[[np.ndarray], npt.ArrayLike]

_CONTACT_SETUP_CACHE: dict[tuple, "ContactSurfaceSpace"] = {}


@dataclass(frozen=True)
class ContactOperators:
    """Container for assembled contact operators."""

    method: str
    coupling_aa: Any | None = None
    coupling_ab: Any | None = None
    B_a: Any | None = None
    B_b: Any | None = None
    B: Any | None = None
    Kuu: Any | None = None
    residual: Any | None = None
    jacobian: Any | None = None
    facet_conn_master: np.ndarray | None = None


def assemble_contact_interface_residual(*args, **kwargs):
    """Assemble residual on a contact interface supermesh."""
    return _assemble_contact_interface_residual(*args, **kwargs)


def assemble_contact_interface_jacobian(*args, **kwargs):
    """Assemble Jacobian on a contact interface supermesh."""
    return _assemble_contact_interface_jacobian(*args, **kwargs)


def assemble_contact_coupling_matrices(*args, **kwargs):
    """Assemble coupling matrices for contact interface constraints."""
    return _assemble_contact_coupling_matrices(*args, **kwargs)


def _coo_to_dense(rows: np.ndarray, cols: np.ndarray, data: np.ndarray, shape: tuple[int, int], *, backend: str):
    if backend == "jax":
        import jax.numpy as jnp

        out = jnp.zeros(shape, dtype=jnp.asarray(data).dtype if np.asarray(data).size else float)
        if len(rows) == 0:
            return out
        return out.at[np.asarray(rows, dtype=int), np.asarray(cols, dtype=int)].add(np.asarray(data))
    out = np.zeros(shape, dtype=float)
    for r, c, v in zip(np.asarray(rows, dtype=int), np.asarray(cols, dtype=int), np.asarray(data, dtype=float)):
        out[int(r), int(c)] += float(v)
    return out


def _p0_reduction_matrix_from_facets(facet_conn: np.ndarray, n_nodes: int):
    facets = np.asarray(facet_conn, dtype=int)
    S = np.zeros((int(facets.shape[0]), int(n_nodes)), dtype=float)
    for f, nodes in enumerate(facets):
        S[int(f), np.asarray(nodes, dtype=int)] = 1.0
    return S


def _coalesce_int_coo(rows: np.ndarray, cols: np.ndarray, data: np.ndarray):
    from ..solver.sparse import coalesce_coo

    r, c, d = coalesce_coo(rows, cols, data)
    return np.asarray(r, dtype=int), np.asarray(c, dtype=int), np.asarray(d, dtype=float)


def _kkt_coo_from_coupling(
    coupling_aa,
    coupling_ab,
    *,
    rho: float,
    multiplier_space: str,
    facet_conn_master: np.ndarray | None,
):
    rows_aa, cols_aa, data_aa = _coalesce_int_coo(coupling_aa.rows, coupling_aa.cols, coupling_aa.data)
    rows_ab, cols_ab, data_ab = _coalesce_int_coo(coupling_ab.rows, coupling_ab.cols, coupling_ab.data)
    n_a = int(coupling_aa.shape[0])
    n_b = int(coupling_ab.shape[1])
    n_u = n_a + n_b

    if multiplier_space == "nodal":
        n_l = n_a
        b_rows = np.concatenate([rows_aa, rows_ab])
        b_cols = np.concatenate([cols_aa, n_a + cols_ab])
        b_data = np.concatenate([data_aa, -data_ab])
    elif multiplier_space == "p0":
        if facet_conn_master is None:
            raise ValueError("facet_conn_master is required when multiplier_space='p0'.")
        facets = np.asarray(facet_conn_master, dtype=int)
        n_l = int(facets.shape[0])
        row_map: dict[int, list[int]] = {i: [] for i in range(n_a)}
        for k, r in enumerate(rows_aa):
            row_map[int(r)].append(int(k))
        row_map_ab: dict[int, list[int]] = {i: [] for i in range(n_a)}
        for k, r in enumerate(rows_ab):
            row_map_ab[int(r)].append(int(k))
        b_rows_l: list[int] = []
        b_cols_l: list[int] = []
        b_data_l: list[float] = []
        for lf, nodes in enumerate(facets):
            acc: dict[int, float] = {}
            for n in np.asarray(nodes, dtype=int):
                for k in row_map.get(int(n), []):
                    c = int(cols_aa[k])
                    acc[c] = acc.get(c, 0.0) + float(data_aa[k])
                for k in row_map_ab.get(int(n), []):
                    c = n_a + int(cols_ab[k])
                    acc[c] = acc.get(c, 0.0) - float(data_ab[k])
            for c, v in acc.items():
                b_rows_l.append(int(lf))
                b_cols_l.append(int(c))
                b_data_l.append(float(v))
        if b_rows_l:
            b_rows = np.asarray(b_rows_l, dtype=int)
            b_cols = np.asarray(b_cols_l, dtype=int)
            b_data = np.asarray(b_data_l, dtype=float)
            b_rows, b_cols, b_data = _coalesce_int_coo(b_rows, b_cols, b_data)
        else:
            b_rows = np.zeros((0,), dtype=int)
            b_cols = np.zeros((0,), dtype=int)
            b_data = np.zeros((0,), dtype=float)
    else:
        raise ValueError("multiplier_space must be 'nodal' or 'p0'")

    # Build Kuu = rho * B^T B from row-wise products.
    by_row: dict[int, list[int]] = {}
    for k, r in enumerate(b_rows):
        by_row.setdefault(int(r), []).append(int(k))
    kuu_acc: dict[tuple[int, int], float] = {}
    if float(rho) != 0.0:
        rr = float(rho)
        for ids in by_row.values():
            for i in ids:
                ci = int(b_cols[i])
                vi = float(b_data[i])
                for j in ids:
                    cj = int(b_cols[j])
                    vj = float(b_data[j])
                    key = (ci, cj)
                    kuu_acc[key] = kuu_acc.get(key, 0.0) + rr * vi * vj

    kuu_rows = np.fromiter((k[0] for k in kuu_acc.keys()), dtype=int, count=len(kuu_acc)) if kuu_acc else np.zeros((0,), dtype=int)
    kuu_cols = np.fromiter((k[1] for k in kuu_acc.keys()), dtype=int, count=len(kuu_acc)) if kuu_acc else np.zeros((0,), dtype=int)
    kuu_data = np.fromiter((v for v in kuu_acc.values()), dtype=float, count=len(kuu_acc)) if kuu_acc else np.zeros((0,), dtype=float)

    # KKT COO assembly:
    # [Kuu  B^T]
    # [ B    0 ]
    k_rows = []
    k_cols = []
    k_data = []
    if kuu_rows.size:
        k_rows.append(kuu_rows)
        k_cols.append(kuu_cols)
        k_data.append(kuu_data)
    if b_rows.size:
        # B^T block (top-right)
        k_rows.append(b_cols)
        k_cols.append(n_u + b_rows)
        k_data.append(b_data)
        # B block (bottom-left)
        k_rows.append(n_u + b_rows)
        k_cols.append(b_cols)
        k_data.append(b_data)
    if k_rows:
        rows = np.concatenate(k_rows)
        cols = np.concatenate(k_cols)
        data = np.concatenate(k_data)
        rows, cols, data = _coalesce_int_coo(rows, cols, data)
    else:
        rows = np.zeros((0,), dtype=int)
        cols = np.zeros((0,), dtype=int)
        data = np.zeros((0,), dtype=float)
    n_total = int(n_u + n_l)
    return rows, cols, data, n_total


def assemble_contact_operators(
    contact,
    *,
    method: str = "mortar",
    rho: float = 0.0,
    multiplier_space: str = "nodal",
    backend: str = "numpy",
    res_form: MixedSurfaceResidualForm | None = None,
    u: Mapping[str, npt.ArrayLike] | Sequence[Any] | None = None,
    params: "WeakParams" | None = None,
    normal_source: str = "master",
    sparse: bool = False,
    batch_jac: bool | None = None,
) -> ContactOperators:
    """
    Assemble contact operators, separating operator build from KKT solve assembly.

    - method='mortar': assemble coupling operators and dense B/Kuu blocks.
    - method='nitsche': assemble residual and Jacobian from a user weak form.
    """
    if method not in {"mortar", "nitsche"}:
        raise ValueError("method must be 'mortar' or 'nitsche'")
    if backend not in {"numpy", "jax"}:
        raise ValueError("backend must be 'numpy' or 'jax'")

    if method == "mortar":
        if not hasattr(contact, "assemble_contact_coupling_matrices"):
            raise TypeError("contact must provide assemble_contact_coupling_matrices() for method='mortar'.")
        coupling_aa, coupling_ab = contact.assemble_contact_coupling_matrices()

        facet_conn_master = None
        if hasattr(contact, "surface_master"):
            facet_conn_master = np.asarray(contact.surface_master.conn, dtype=int)
        elif hasattr(contact, "contacts") and len(getattr(contact, "contacts")) > 0:
            first = contact.contacts[0]
            if hasattr(first, "surface_master"):
                facet_conn_master = np.asarray(first.surface_master.conn, dtype=int)

        M_aa = _coo_to_dense(
            coupling_aa.rows, coupling_aa.cols, coupling_aa.data, coupling_aa.shape, backend=backend
        )
        M_ab = _coo_to_dense(
            coupling_ab.rows, coupling_ab.cols, coupling_ab.data, coupling_ab.shape, backend=backend
        )

        if backend == "jax":
            import jax.numpy as jnp

            xp = jnp
        else:
            xp = np

        if multiplier_space == "nodal":
            B_a = M_aa
            B_b = M_ab
        elif multiplier_space == "p0":
            if facet_conn_master is None:
                raise ValueError("facet_conn_master is required when multiplier_space='p0'.")
            n_master_nodes = int(coupling_aa.shape[0])
            S_np = _p0_reduction_matrix_from_facets(facet_conn_master, n_master_nodes)
            S = xp.asarray(S_np)
            B_a = S @ M_aa
            B_b = S @ M_ab
        else:
            raise ValueError("multiplier_space must be 'nodal' or 'p0'")

        B = xp.concatenate([B_a, -B_b], axis=1)
        Kuu = xp.asarray(rho) * (B.T @ B)
        return ContactOperators(
            method="mortar",
            coupling_aa=coupling_aa,
            coupling_ab=coupling_ab,
            B_a=B_a,
            B_b=B_b,
            B=B,
            Kuu=Kuu,
            facet_conn_master=facet_conn_master,
        )

    if res_form is None or u is None or params is None:
        raise ValueError("method='nitsche' requires res_form, u, and params.")
    if not hasattr(contact, "assemble_residual") or not hasattr(contact, "assemble_jacobian"):
        raise TypeError("contact must provide assemble_residual() and assemble_jacobian() for method='nitsche'.")
    residual = contact.assemble_residual(res_form, u, params, normal_source=normal_source)
    jacobian = contact.assemble_jacobian(
        res_form,
        u,
        params,
        normal_source=normal_source,
        sparse=sparse,
        backend=backend,
        batch_jac=batch_jac,
    )
    return ContactOperators(
        method="nitsche",
        residual=residual,
        jacobian=jacobian,
    )


def assemble_contact_kkt(
    coupling_aa,
    coupling_ab,
    *,
    rho: float = 0.0,
    multiplier_space: str = "nodal",
    facet_conn_master: np.ndarray | None = None,
    backend: str = "numpy",
    format: str = "fluxsparse",
    return_blocks: bool = False,
):
    """
    Assemble contact KKT block from coupling matrices.

    KKT is assembled as:
      B = [B_a, -B_b]
      Kuu = rho * (B^T B)
      KKT = [[Kuu, B^T], [B, 0]]

    multiplier_space:
    - "nodal": lambda lives on interface nodal basis (B_a=M_aa, B_b=M_ab)
    - "p0": lambda is facet-wise constant on master side (B_* = S * M_*)
    """
    if backend not in {"numpy", "jax"}:
        raise ValueError("backend must be 'numpy' or 'jax'")
    if multiplier_space not in {"nodal", "p0"}:
        raise ValueError("multiplier_space must be 'nodal' or 'p0'")
    if format not in {"dense", "fluxsparse", "bcoo"}:
        raise ValueError("format must be 'dense', 'fluxsparse', or 'bcoo'")
    if return_blocks and format != "dense":
        raise ValueError("return_blocks=True is supported only with format='dense'.")

    if format != "dense":
        import jax

        if isinstance(rho, jax.core.Tracer):
            raise ValueError("format='fluxsparse'/'bcoo' currently requires rho to be a concrete scalar.")
        rows, cols, data, n_total = _kkt_coo_from_coupling(
            coupling_aa,
            coupling_ab,
            rho=float(rho),
            multiplier_space=multiplier_space,
            facet_conn_master=facet_conn_master,
        )
        if format == "fluxsparse":
            from ..solver import FluxSparseMatrix

            return FluxSparseMatrix(rows, cols, data, n_dofs=n_total)
        from jax.experimental import sparse as jsparse
        import jax.numpy as jnp

        idx = jnp.stack([jnp.asarray(rows, dtype=jnp.int32), jnp.asarray(cols, dtype=jnp.int32)], axis=-1)
        return jsparse.BCOO((jnp.asarray(data), idx), shape=(n_total, n_total))

    M_aa = _coo_to_dense(coupling_aa.rows, coupling_aa.cols, coupling_aa.data, coupling_aa.shape, backend=backend)
    M_ab = _coo_to_dense(coupling_ab.rows, coupling_ab.cols, coupling_ab.data, coupling_ab.shape, backend=backend)

    if backend == "jax":
        import jax.numpy as jnp

        xp = jnp
    else:
        xp = np

    if multiplier_space == "nodal":
        B_a = M_aa
        B_b = M_ab
    else:
        if facet_conn_master is None:
            raise ValueError("facet_conn_master is required when multiplier_space='p0'.")
        n_master_nodes = int(coupling_aa.shape[0])
        S_np = _p0_reduction_matrix_from_facets(facet_conn_master, n_master_nodes)
        S = xp.asarray(S_np)
        B_a = S @ M_aa
        B_b = S @ M_ab

    B = xp.concatenate([B_a, -B_b], axis=1)
    Kuu = xp.asarray(rho) * (B.T @ B)
    n_lambda = int(B.shape[0])
    Zll = xp.zeros((n_lambda, n_lambda), dtype=Kuu.dtype)
    KKT = xp.block([[Kuu, B.T], [B, Zll]])

    if return_blocks:
        return KKT, B_a, B_b
    return KKT


def solve_contact_kkt(kkt_matrix, rhs, *, backend: str = "numpy", diagonal_shift: float = 0.0):
    """
    Solve KKT linear system KKT * x = rhs.

    backend:
    - "numpy": dense numpy solve
    - "jax": custom_vjp-enabled dense solve with implicit differentiation
    """
    if backend == "numpy":
        if hasattr(kkt_matrix, "to_dense"):
            A = np.asarray(kkt_matrix.to_dense(), dtype=float)
        else:
            A = np.asarray(kkt_matrix, dtype=float)
        if diagonal_shift != 0.0:
            A = A + float(diagonal_shift) * np.eye(A.shape[0], dtype=A.dtype)
        return np.linalg.solve(A, np.asarray(rhs, dtype=float))
    if backend != "jax":
        raise ValueError("backend must be 'numpy' or 'jax'")

    import jax
    import jax.numpy as jnp

    @jax.custom_vjp
    def _solve_jax(A, b):
        return jnp.linalg.solve(A, b)

    def _solve_jax_fwd(A, b):
        x = jnp.linalg.solve(A, b)
        return x, (A, x)

    def _solve_jax_bwd(res, g):
        A, x = res
        lam = jnp.linalg.solve(A.T, g)
        gA = -jnp.outer(lam, x)
        gb = lam
        return gA, gb

    _solve_jax.defvjp(_solve_jax_fwd, _solve_jax_bwd)
    if hasattr(kkt_matrix, "to_dense"):
        A = jnp.asarray(kkt_matrix.to_dense())
    else:
        try:
            from jax.experimental import sparse as jsparse  # type: ignore
        except Exception:
            jsparse = None
        if jsparse is not None and isinstance(kkt_matrix, jsparse.BCOO):
            A = kkt_matrix.todense()
        else:
            A = jnp.asarray(kkt_matrix)
    if diagonal_shift != 0.0:
        A = A + jnp.asarray(diagonal_shift, dtype=A.dtype) * jnp.eye(A.shape[0], dtype=A.dtype)
    return _solve_jax(A, jnp.asarray(rhs))


@dataclass(frozen=True)
class ContactSide:
    surface: SurfaceMesh
    elem_conn: np.ndarray | None
    value_dim: int
    space: object | None = None

    @classmethod
    def from_facets(
        cls,
        mesh: BaseMesh,
        facets: np.ndarray,
        space=None,
        *,
        value_dim: int | None = None,
        mode: str = "touching",
    ):
        side = mesh.surface_with_elem_conn_from_facets(facets, mode=mode)
        if value_dim is None:
            if space is None:
                raise ValueError("space or value_dim is required for ContactSide.from_facets")
            value_dim = int(getattr(space, "value_dim", 1))
        return cls(surface=side.surface, elem_conn=side.elem_conn, value_dim=int(value_dim), space=space)

    @classmethod
    def from_surfaces(
        cls,
        surface: SurfaceMesh,
        *,
        elem_conn: np.ndarray | None = None,
        value_dim: int = 1,
        space: object | None = None,
    ):
        return cls(surface=surface, elem_conn=elem_conn, value_dim=int(value_dim), space=space)


def _facet_map_for_elem_conn(surface: SurfaceMesh, elem_conn: np.ndarray | None) -> np.ndarray:
    if elem_conn is None:
        raise ValueError("elem_conn is required to build facet_to_elem mapping.")
    if elem_conn.shape[1] in {4, 10}:
        return map_surface_facets_to_tet_elements(surface, elem_conn)
    if elem_conn.shape[1] in {8, 20, 27}:
        return map_surface_facets_to_hex_elements(surface, elem_conn)
    raise NotImplementedError("elem_conn must be tet4/tet10/hex8/hex20/hex27")


def facet_gap_values(
    coords: np.ndarray,
    facets: np.ndarray,
    u: np.ndarray,
    n: np.ndarray,
    c: float,
    *,
    value_dim: int | None = None,
    reduce: str = "min",
) -> tuple[np.ndarray, float]:
    """
    Compute per-facet gap values for a one-sided contact plane.

    Returns (g_f, min_g_all) where g_f is reduced per facet and min_g_all is
    the global minimum node gap.
    """
    coords_np = np.asarray(coords, dtype=float)
    if value_dim is None:
        value_dim = int(coords_np.shape[1])
    u_nodes = np.asarray(u, dtype=float).reshape(-1, value_dim)
    x_cur = coords_np + u_nodes
    g_all = np.dot(x_cur, np.asarray(n, dtype=float)) - float(c)
    min_g_all = float(np.min(g_all)) if g_all.size else 0.0
    if facets is None or len(facets) == 0:
        return np.zeros((0,), dtype=float), min_g_all
    if reduce == "min":
        g_f = np.array([np.min(g_all[np.asarray(facet, dtype=int)]) for facet in facets], dtype=float)
    elif reduce == "mean":
        g_f = np.array([np.mean(g_all[np.asarray(facet, dtype=int)]) for facet in facets], dtype=float)
    else:
        raise ValueError("reduce must be 'min' or 'mean'")
    return g_f, min_g_all


def active_contact_facets(
    coords: np.ndarray,
    facets: np.ndarray,
    u: np.ndarray,
    n: np.ndarray,
    c: float,
    *,
    value_dim: int | None = None,
    reduce: str = "min",
    threshold: float = 0.0,
) -> tuple[np.ndarray, float]:
    """Return active facet indices and global minimum gap for one-sided contact."""
    g_f, min_g_all = facet_gap_values(
        coords,
        facets,
        u,
        n,
        c,
        value_dim=value_dim,
        reduce=reduce,
    )
    active_ids = np.nonzero(g_f < threshold)[0]
    return active_ids, min_g_all


@dataclass(frozen=True)
class OneSidedContact:
    side: ContactSide
    n: np.ndarray | None
    c: float
    k: float
    beta: float
    quad_order: int = 2
    normal_sign: float = 1.0
    tol: float = 1e-8
    facet_map: np.ndarray | None = None

    @classmethod
    def from_side(
        cls,
        side: ContactSide,
        *,
        n: np.ndarray | None,
        c: float,
        k: float,
        beta: float,
        quad_order: int = 2,
        normal_sign: float = 1.0,
        tol: float = 1e-8,
        facet_map: np.ndarray | None = None,
    ) -> "OneSidedContact":
        if facet_map is None:
            facet_map = _facet_map_for_elem_conn(side.surface, side.elem_conn)
        return cls(
            side=side,
            n=n,
            c=float(c),
            k=float(k),
            beta=float(beta),
            quad_order=int(quad_order),
            normal_sign=float(normal_sign),
            tol=float(tol),
            facet_map=facet_map,
        )

    def assemble(self, u, *, return_metrics: bool = False):
        return assemble_contact_onesided_floor(
            self.side.surface,
            np.asarray(u, dtype=float),
            n=None if self.n is None else np.asarray(self.n, dtype=float),
            c=self.c,
            k=self.k,
            beta=self.beta,
            value_dim=self.side.value_dim,
            elem_conn=np.asarray(self.side.elem_conn) if self.side.elem_conn is not None else None,
            facet_to_elem=self.facet_map,
            quad_order=self.quad_order,
            normal_sign=self.normal_sign,
            tol=self.tol,
            return_metrics=return_metrics,
        )


@dataclass(eq=False)
class OneSidedContactSurfaceSpace:
    """Surface wrapper for one-sided (Dirichlet) contact assembly."""

    surface_slave: SurfaceMesh
    elem_conn_slave: np.ndarray
    facet_to_elem_slave: np.ndarray
    value_dim: int = 1
    quad_order: int = 2
    normal_sign: float = 1.0
    tol: float = 1e-8
    surface_master: SurfaceMesh | None = None
    elem_conn_master: np.ndarray | None = None
    facet_to_elem_master: np.ndarray | None = None

    @classmethod
    def from_side(
        cls,
        side: ContactSide,
        *,
        surface_master: SurfaceMesh | None = None,
        elem_conn_master: np.ndarray | None = None,
        facet_to_elem_master: np.ndarray | None = None,
        quad_order: int = 2,
        normal_sign: float = 1.0,
        tol: float = 1e-8,
    ) -> "OneSidedContactSurfaceSpace":
        if side.elem_conn is None:
            raise ValueError("side.elem_conn is required for one-sided assembly")
        facet_map_slave = _facet_map_for_elem_conn(side.surface, side.elem_conn)
        facet_map_master = facet_to_elem_master
        if surface_master is not None and elem_conn_master is not None and facet_map_master is None:
            facet_map_master = _facet_map_for_elem_conn(surface_master, elem_conn_master)
        return cls(
            surface_slave=side.surface,
            elem_conn_slave=np.asarray(side.elem_conn, dtype=int),
            facet_to_elem_slave=np.asarray(facet_map_slave, dtype=int),
            value_dim=int(side.value_dim),
            quad_order=int(quad_order),
            normal_sign=float(normal_sign),
            tol=float(tol),
            surface_master=surface_master,
            elem_conn_master=None if elem_conn_master is None else np.asarray(elem_conn_master, dtype=int),
            facet_to_elem_master=None if facet_map_master is None else np.asarray(facet_map_master, dtype=int),
        )

    @classmethod
    def from_facets(
        cls,
        mesh: BaseMesh,
        facets: np.ndarray,
        space=None,
        *,
        surface_master: SurfaceMesh | None = None,
        elem_conn_master: np.ndarray | None = None,
        facet_to_elem_master: np.ndarray | None = None,
        value_dim: int | None = None,
        quad_order: int = 2,
        normal_sign: float = 1.0,
        tol: float = 1e-8,
        mode: str = "touching",
    ) -> "OneSidedContactSurfaceSpace":
        side = ContactSide.from_facets(mesh, facets, space, value_dim=value_dim, mode=mode)
        return cls.from_side(
            side,
            surface_master=surface_master,
            elem_conn_master=elem_conn_master,
            facet_to_elem_master=facet_to_elem_master,
            quad_order=quad_order,
            normal_sign=normal_sign,
            tol=tol,
        )

    def assemble_bilinear(
        self,
        u_hat_fn: SurfaceHatFn | None,
        params: "WeakParams",
        *,
        u_master: np.ndarray | None = None,
        grad_source: str = "volume",
        dof_source: str = "volume",
        quad_order: int | None = None,
        normal_sign: float | None = None,
        tol: float | None = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        return assemble_onesided_bilinear(
            self.surface_slave,
            u_hat_fn,
            params,
            surface_master=self.surface_master,
            u_master=u_master,
            value_dim=self.value_dim,
            elem_conn=self.elem_conn_slave,
            facet_to_elem=self.facet_to_elem_slave,
            elem_conn_master=self.elem_conn_master,
            facet_to_elem_master=self.facet_to_elem_master,
            grad_source=grad_source,
            dof_source=dof_source,
            quad_order=self.quad_order if quad_order is None else int(quad_order),
            normal_sign=self.normal_sign if normal_sign is None else float(normal_sign),
            tol=self.tol if tol is None else float(tol),
        )


@dataclass(eq=False)
class ContactSurfaceSpace:
    """Surface interface wrapper for contact assembly on a supermesh."""

    surface_master: SurfaceMesh
    surface_slave: SurfaceMesh
    supermesh_coords: np.ndarray
    supermesh_conn: np.ndarray
    source_facets_master: np.ndarray
    source_facets_slave: np.ndarray
    elem_conn_master: np.ndarray | None
    elem_conn_slave: np.ndarray | None
    facet_to_elem_master: np.ndarray | None
    facet_to_elem_slave: np.ndarray | None
    field_master: str = "a"
    field_slave: str = "b"
    value_dim_master: int = 1
    value_dim_slave: int = 1
    space_mode_master: str = "nodal"
    space_mode_slave: str = "nodal"
    facet_dofs_master: np.ndarray | None = None
    facet_dofs_slave: np.ndarray | None = None
    quad_order: int = 1
    normal_sign: float | None = None
    tol: float = 1e-8
    backend: str = "jax"
    fd_eps: float = 1e-6
    fd_mode: str = "central"
    fd_block_size: int = 1
    batch_jac: bool | None = None

    @classmethod
    def from_surfaces(
        cls,
        surface_master: SurfaceMesh,
        surface_slave: SurfaceMesh,
        *,
        elem_conn_master: np.ndarray | None = None,
        elem_conn_slave: np.ndarray | None = None,
        field_master: str = "a",
        field_slave: str = "b",
        value_dim_master: int = 1,
        value_dim_slave: int = 1,
        space_mode_master: str = "nodal",
        space_mode_slave: str = "nodal",
        facet_dofs_master: np.ndarray | None = None,
        facet_dofs_slave: np.ndarray | None = None,
        quad_order: int = 1,
        normal_sign: float | None = None,
        tol: float = 1e-8,
        backend: str = "jax",
        fd_eps: float = 1e-6,
        fd_mode: str = "central",
        fd_block_size: int = 1,
        batch_jac: bool | None = None,
        setup_cache_enabled: bool | None = None,
        setup_cache_trace: bool | None = None,
    ) -> "ContactSurfaceSpace":
        import hashlib
        import os

        if setup_cache_enabled is None:
            setup_cache_enabled = os.getenv("FLUXFEM_CONTACT_SETUP_CACHE", "0") not in ("0", "", "false", "False")
        if setup_cache_trace is None:
            setup_cache_trace = os.getenv("FLUXFEM_CONTACT_SETUP_CACHE_TRACE", "0") not in ("0", "", "false", "False")

        def _array_sig(arr: np.ndarray) -> tuple:
            arr_c = np.ascontiguousarray(arr)
            h = hashlib.blake2b(arr_c.view(np.uint8), digest_size=8).hexdigest()
            return (arr_c.shape, str(arr_c.dtype), h)

        if setup_cache_enabled:
            global _CONTACT_SETUP_CACHE
            try:
                _CONTACT_SETUP_CACHE
            except NameError:
                _CONTACT_SETUP_CACHE = {}
            key = (
                _array_sig(np.asarray(surface_master.coords)),
                _array_sig(np.asarray(surface_master.conn)),
                _array_sig(np.asarray(surface_slave.coords)),
                _array_sig(np.asarray(surface_slave.conn)),
                None if elem_conn_master is None else _array_sig(np.asarray(elem_conn_master)),
                None if elem_conn_slave is None else _array_sig(np.asarray(elem_conn_slave)),
                field_master,
                field_slave,
                int(value_dim_master),
                int(value_dim_slave),
                str(space_mode_master),
                str(space_mode_slave),
                None if facet_dofs_master is None else _array_sig(np.asarray(facet_dofs_master)),
                None if facet_dofs_slave is None else _array_sig(np.asarray(facet_dofs_slave)),
                int(quad_order),
                float(normal_sign) if normal_sign is not None else None,
                float(tol),
                backend,
                float(fd_eps),
                fd_mode,
                int(fd_block_size),
                bool(batch_jac) if batch_jac is not None else None,
            )
            cached = _CONTACT_SETUP_CACHE.get(key)
            if cached is not None:
                if setup_cache_trace:
                    print(
                        f"[contact] setup cache hit n_tris={int(cached.supermesh_conn.shape[0])}",
                        flush=True,
                    )
                return cached

        sm = build_surface_supermesh(surface_master, surface_slave, tol=tol)
        facet_map_master = None
        facet_map_slave = None
        if elem_conn_master is not None:
            if elem_conn_master.shape[1] in {4, 10}:
                facet_map_master = map_surface_facets_to_tet_elements(surface_master, elem_conn_master)
            elif elem_conn_master.shape[1] in {8, 20, 27}:
                facet_map_master = map_surface_facets_to_hex_elements(surface_master, elem_conn_master)
            else:
                raise NotImplementedError("elem_conn_master must be tet4/tet10/hex8/hex20/hex27")
        if elem_conn_slave is not None:
            if elem_conn_slave.shape[1] in {4, 10}:
                facet_map_slave = map_surface_facets_to_tet_elements(surface_slave, elem_conn_slave)
            elif elem_conn_slave.shape[1] in {8, 20, 27}:
                facet_map_slave = map_surface_facets_to_hex_elements(surface_slave, elem_conn_slave)
            else:
                raise NotImplementedError("elem_conn_slave must be tet4/tet10/hex8/hex20/hex27")
        obj = cls(
            surface_master=surface_master,
            surface_slave=surface_slave,
            supermesh_coords=sm.coords,
            supermesh_conn=sm.conn,
            source_facets_master=sm.source_facets_a,
            source_facets_slave=sm.source_facets_b,
            elem_conn_master=elem_conn_master,
            elem_conn_slave=elem_conn_slave,
            facet_to_elem_master=facet_map_master,
            facet_to_elem_slave=facet_map_slave,
            field_master=field_master,
            field_slave=field_slave,
            value_dim_master=value_dim_master,
            value_dim_slave=value_dim_slave,
            space_mode_master=space_mode_master,
            space_mode_slave=space_mode_slave,
            facet_dofs_master=None if facet_dofs_master is None else np.asarray(facet_dofs_master, dtype=int),
            facet_dofs_slave=None if facet_dofs_slave is None else np.asarray(facet_dofs_slave, dtype=int),
            quad_order=quad_order,
            normal_sign=normal_sign,
            tol=tol,
            backend=backend,
            fd_eps=fd_eps,
            fd_mode=fd_mode,
            fd_block_size=fd_block_size,
            batch_jac=batch_jac,
        )
        if setup_cache_enabled:
            _CONTACT_SETUP_CACHE[key] = obj
            if setup_cache_trace:
                print(
                    f"[contact] setup cache store n_tris={int(obj.supermesh_conn.shape[0])}",
                    flush=True,
                )
        return obj

    @classmethod
    def from_facets(
        cls,
        coords: np.ndarray,
        facets: np.ndarray,
        *,
        elem_conn: np.ndarray | None = None,
        value_dim: int = 1,
        space_mode: str = "nodal",
        facet_dofs: np.ndarray | None = None,
        quad_order: int = 1,
        normal_sign: float | None = None,
        tol: float = 1e-8,
        backend: str = "jax",
        fd_eps: float = 1e-6,
        fd_mode: str = "central",
        fd_block_size: int = 1,
        batch_jac: bool | None = None,
        setup_cache_enabled: bool | None = None,
        setup_cache_trace: bool | None = None,
    ) -> "ContactSurfaceSpace":
        surface = SurfaceMesh.from_facets(coords, facets)
        return cls.from_surfaces(
            surface,
            surface,
            elem_conn_master=elem_conn,
            elem_conn_slave=elem_conn,
            value_dim_master=value_dim,
            value_dim_slave=value_dim,
            space_mode_master=space_mode,
            space_mode_slave=space_mode,
            facet_dofs_master=facet_dofs,
            facet_dofs_slave=facet_dofs,
            quad_order=quad_order,
            normal_sign=normal_sign,
            tol=tol,
            backend=backend,
            fd_eps=fd_eps,
            fd_mode=fd_mode,
            fd_block_size=fd_block_size,
            batch_jac=batch_jac,
            setup_cache_enabled=setup_cache_enabled,
            setup_cache_trace=setup_cache_trace,
        )

    @classmethod
    def from_surfaces_and_spaces(
        cls,
        surface_master: SurfaceMesh,
        surface_slave: SurfaceMesh,
        space_master,
        space_slave,
        *,
        elem_conn_master: np.ndarray | None = None,
        elem_conn_slave: np.ndarray | None = None,
        field_master: str = "a",
        field_slave: str = "b",
        value_dim_master: int | None = None,
        value_dim_slave: int | None = None,
        space_mode_master: str = "nodal",
        space_mode_slave: str = "nodal",
        facet_dofs_master: np.ndarray | None = None,
        facet_dofs_slave: np.ndarray | None = None,
        quad_order: int = 1,
        normal_sign: float | None = None,
        tol: float = 1e-8,
        backend: str = "jax",
        fd_eps: float = 1e-6,
        fd_mode: str = "central",
        fd_block_size: int = 1,
        batch_jac: bool | None = None,
    ) -> "ContactSurfaceSpace":
        if value_dim_master is None:
            value_dim_master = int(getattr(space_master, "value_dim", 1))
        if value_dim_slave is None:
            value_dim_slave = int(getattr(space_slave, "value_dim", 1))
        return cls.from_surfaces(
            surface_master,
            surface_slave,
            elem_conn_master=elem_conn_master,
            elem_conn_slave=elem_conn_slave,
            field_master=field_master,
            field_slave=field_slave,
            value_dim_master=value_dim_master,
            value_dim_slave=value_dim_slave,
            space_mode_master=space_mode_master,
            space_mode_slave=space_mode_slave,
            facet_dofs_master=facet_dofs_master,
            facet_dofs_slave=facet_dofs_slave,
            quad_order=quad_order,
            normal_sign=normal_sign,
            tol=tol,
            backend=backend,
            fd_eps=fd_eps,
            fd_mode=fd_mode,
            fd_block_size=fd_block_size,
            batch_jac=batch_jac,
        )

    @classmethod
    def from_sides(
        cls,
        master: ContactSide,
        slave: ContactSide,
        *,
        field_master: str = "a",
        field_slave: str = "b",
        quad_order: int = 1,
        space_mode_master: str = "nodal",
        space_mode_slave: str = "nodal",
        facet_dofs_master: np.ndarray | None = None,
        facet_dofs_slave: np.ndarray | None = None,
        normal_sign: float | None = None,
        tol: float = 1e-8,
        backend: str = "jax",
        fd_eps: float = 1e-6,
        fd_mode: str = "central",
        fd_block_size: int = 1,
        batch_jac: bool | None = None,
        setup_cache_enabled: bool | None = None,
        setup_cache_trace: bool | None = None,
    ) -> "ContactSurfaceSpace":
        return cls.from_surfaces(
            master.surface,
            slave.surface,
            elem_conn_master=master.elem_conn,
            elem_conn_slave=slave.elem_conn,
            field_master=field_master,
            field_slave=field_slave,
            value_dim_master=master.value_dim,
            value_dim_slave=slave.value_dim,
            space_mode_master=space_mode_master,
            space_mode_slave=space_mode_slave,
            facet_dofs_master=facet_dofs_master,
            facet_dofs_slave=facet_dofs_slave,
            quad_order=quad_order,
            normal_sign=normal_sign,
            tol=tol,
            backend=backend,
            fd_eps=fd_eps,
            fd_mode=fd_mode,
            fd_block_size=fd_block_size,
            batch_jac=batch_jac,
            setup_cache_enabled=setup_cache_enabled,
            setup_cache_trace=setup_cache_trace,
        )

    @classmethod  # type: ignore[no-redef]
    def from_facets(
        cls,
        coords_master: np.ndarray,
        facets_master: np.ndarray,
        coords_slave: np.ndarray,
        facets_slave: np.ndarray,
        *,
        elem_conn_master: np.ndarray | None = None,
        elem_conn_slave: np.ndarray | None = None,
        field_master: str = "a",
        field_slave: str = "b",
        value_dim_master: int = 1,
        value_dim_slave: int = 1,
        space_mode_master: str = "nodal",
        space_mode_slave: str = "nodal",
        facet_dofs_master: np.ndarray | None = None,
        facet_dofs_slave: np.ndarray | None = None,
        quad_order: int = 1,
        normal_sign: float | None = None,
        tol: float = 1e-8,
        backend: str = "jax",
        fd_eps: float = 1e-6,
        fd_mode: str = "central",
        batch_jac: bool | None = None,
        setup_cache_enabled: bool | None = None,
        setup_cache_trace: bool | None = None,
    ) -> "ContactSurfaceSpace":
        surface_master = SurfaceMesh.from_facets(coords_master, facets_master)
        surface_slave = SurfaceMesh.from_facets(coords_slave, facets_slave)
        return cls.from_surfaces(
            surface_master,
            surface_slave,
            elem_conn_master=elem_conn_master,
            elem_conn_slave=elem_conn_slave,
            field_master=field_master,
            field_slave=field_slave,
            value_dim_master=value_dim_master,
            value_dim_slave=value_dim_slave,
            space_mode_master=space_mode_master,
            space_mode_slave=space_mode_slave,
            facet_dofs_master=facet_dofs_master,
            facet_dofs_slave=facet_dofs_slave,
            quad_order=quad_order,
            normal_sign=normal_sign,
            tol=tol,
            backend=backend,
            fd_eps=fd_eps,
            fd_mode=fd_mode,
            batch_jac=batch_jac,
            setup_cache_enabled=setup_cache_enabled,
            setup_cache_trace=setup_cache_trace,
        )

    def _split_fields(self, u: Mapping[str, np.ndarray] | Sequence[np.ndarray]):
        if isinstance(u, Mapping):
            return u[self.field_master], u[self.field_slave]
        if len(u) != 2:
            raise ValueError("u must be a mapping or a length-2 sequence")
        return u[0], u[1]

    def _auto_normal_sign(self) -> float:
        if not hasattr(self.surface_master, "facet_normals"):
            return 1.0
        normals = self.surface_master.facet_normals()
        coords = np.asarray(self.surface_master.coords)
        coords_slave = np.asarray(self.surface_slave.coords)
        facets_m = np.asarray(self.surface_master.conn, dtype=int)
        facets_s = np.asarray(self.surface_slave.conn, dtype=int)
        dots = []
        for fa, fb in zip(self.source_facets_master, self.source_facets_slave):
            n = normals[int(fa)]
            cm = np.mean(coords[facets_m[int(fa)]], axis=0)
            cs = np.mean(coords_slave[facets_s[int(fb)]], axis=0)
            dots.append(float(np.dot(n, cs - cm)))
        if not dots:
            return 1.0
        return 1.0 if np.sum(dots) >= 0.0 else -1.0

    def _resolve_backend(self, backend: str | None) -> str:
        use_backend = self.backend if backend is None else backend
        if use_backend not in {"jax", "numpy"}:
            raise ValueError("backend must be 'jax' or 'numpy'")
        return use_backend

    def assemble_contact_coupling_matrices(self) -> tuple["ContactCouplingMatrix", "ContactCouplingMatrix"]:
        """Return (M_aa, M_ab) coupling matrices on this contact interface."""
        return _assemble_contact_coupling_matrices(
            self.supermesh_coords,
            self.supermesh_conn,
            self.source_facets_master,
            self.source_facets_slave,
            self.surface_master,
            self.surface_slave,
        )

    def assemble_contact_kkt(
        self,
        *,
        rho: float = 0.0,
        multiplier_space: str = "nodal",
        backend: str = "numpy",
        format: str = "fluxsparse",
        return_blocks: bool = False,
    ):
        m_aa, m_ab = self.assemble_contact_coupling_matrices()
        return assemble_contact_kkt(
            m_aa,
            m_ab,
            rho=rho,
            multiplier_space=multiplier_space,
            facet_conn_master=np.asarray(self.surface_master.conn, dtype=int),
            backend=backend,
            format=format,
            return_blocks=return_blocks,
        )

    def assemble_contact_operators(
        self,
        *,
        method: str = "mortar",
        rho: float = 0.0,
        multiplier_space: str = "nodal",
        backend: str = "numpy",
        res_form: MixedSurfaceResidualForm | None = None,
        u: Mapping[str, npt.ArrayLike] | Sequence[npt.ArrayLike] | None = None,
        params: "WeakParams" | None = None,
        normal_source: str = "master",
        sparse: bool = False,
        batch_jac: bool | None = None,
    ) -> ContactOperators:
        return assemble_contact_operators(
            self,
            method=method,
            rho=rho,
            multiplier_space=multiplier_space,
            backend=backend,
            res_form=res_form,
            u=u,
            params=params,
            normal_source=normal_source,
            sparse=sparse,
            batch_jac=batch_jac,
        )

    def assemble_residual(
        self,
        res_form: MixedSurfaceResidualForm,
        u: Mapping[str, npt.ArrayLike] | Sequence[npt.ArrayLike],
        params: "WeakParams",
        *,
        normal_sign: float | None = None,
        normal_source: str = "master",
    ) -> np.ndarray:
        u_master, u_slave = self._split_fields(u)
        if normal_sign is None:
            normal_sign = self.normal_sign
        if normal_sign is None:
            normal_sign = self._auto_normal_sign()
        return _assemble_contact_interface_residual(
            self.supermesh_coords,
            self.supermesh_conn,
            self.source_facets_master,
            self.source_facets_slave,
            self.surface_master,
            self.surface_slave,
            res_form,
            u_master,
            u_slave,
            params,
            value_dim_a=self.value_dim_master,
            value_dim_b=self.value_dim_slave,
            space_mode_a=self.space_mode_master,
            space_mode_b=self.space_mode_slave,
            facet_dofs_a=self.facet_dofs_master,
            facet_dofs_b=self.facet_dofs_slave,
            field_a=self.field_master,
            field_b=self.field_slave,
            elem_conn_a=self.elem_conn_master,
            elem_conn_b=self.elem_conn_slave,
            facet_to_elem_a=self.facet_to_elem_master,
            facet_to_elem_b=self.facet_to_elem_slave,
            normal_source=normal_source,
            normal_from="master",
            master_field=self.field_master,
            normal_sign=normal_sign,
            grad_source="volume",
            dof_source="volume",
            quad_order=self.quad_order,
            tol=self.tol,
        )

    def assemble_jacobian(
        self,
        res_form: MixedSurfaceResidualForm,
        u: Mapping[str, npt.ArrayLike] | Sequence[npt.ArrayLike],
        params: "WeakParams",
        *,
        normal_sign: float | None = None,
        normal_source: str = "master",
        sparse: bool = False,
        backend: str | None = None,
        batch_jac: bool | None = None,
    ) -> ContactJacobianReturn:
        u_master, u_slave = self._split_fields(u)
        if normal_sign is None:
            normal_sign = self.normal_sign
        if normal_sign is None:
            normal_sign = self._auto_normal_sign()
        use_backend = self._resolve_backend(backend)
        use_batch_jac = self.batch_jac if batch_jac is None else batch_jac
        return _assemble_contact_interface_jacobian(
            self.supermesh_coords,
            self.supermesh_conn,
            self.source_facets_master,
            self.source_facets_slave,
            self.surface_master,
            self.surface_slave,
            res_form,
            u_master,
            u_slave,
            params,
            value_dim_a=self.value_dim_master,
            value_dim_b=self.value_dim_slave,
            space_mode_a=self.space_mode_master,
            space_mode_b=self.space_mode_slave,
            facet_dofs_a=self.facet_dofs_master,
            facet_dofs_b=self.facet_dofs_slave,
            field_a=self.field_master,
            field_b=self.field_slave,
            elem_conn_a=self.elem_conn_master,
            elem_conn_b=self.elem_conn_slave,
            facet_to_elem_a=self.facet_to_elem_master,
            facet_to_elem_b=self.facet_to_elem_slave,
            normal_source=normal_source,
            normal_from="master",
            master_field=self.field_master,
            normal_sign=normal_sign,
            grad_source="volume",
            dof_source="volume",
            quad_order=self.quad_order,
            tol=self.tol,
            sparse=sparse,
            backend=use_backend,
            batch_jac=use_batch_jac,
            fd_eps=self.fd_eps,
            fd_mode=self.fd_mode,
            fd_block_size=self.fd_block_size,
        )

    def assemble_bilinear(
        self,
        bilin: Callable[..., Any],
        u_master: Mapping[str, npt.ArrayLike] | Sequence[npt.ArrayLike] | npt.ArrayLike,
        u_slave: npt.ArrayLike | None = None,
        params: "WeakParams" | None = None,
        *,
        sparse: bool = False,
        normal_source: str = "master",
    ) -> ContactJacobianReturn:
        """
        Assemble a mixed surface bilinear form with signature (v1, v2, u1, u2, params).

        Notes:
        - v1/v2/u1/u2 are symbolic field refs; use .val/.grad/.sym_grad in the expression.
        - The bilinear must be linear in v1 and v2 and include ds() in its expression.
        - When building dot products, prefer dot(v1, ...) and dot(v2, ...) to keep shapes consistent.
        - Normal orientation, grad_source, and dof_source are fixed internally for simplicity.
        - u_master/u_slave can be passed as a single mapping/length-2 sequence; in that case,
          pass params as the next positional arg or a keyword.
        """
        from ..core.weakform import (
            compile_mixed_surface_residual,
            compile_mixed_surface_residual_numpy,
            unknown_ref,
            test_ref,
            param_ref,
            zero_ref,
        )

        def _is_field_pair(obj) -> bool:
            if isinstance(obj, Mapping):
                return True
            return isinstance(obj, Sequence) and not hasattr(obj, "shape")

        if params is None:
            if u_slave is None:
                raise TypeError("params is required")
            if _is_field_pair(u_master):
                params = u_slave
                u_master, u_slave = self._split_fields(u_master)
            else:
                raise TypeError("params is required")
        elif u_slave is None:
            u_master, u_slave = self._split_fields(u_master)

        v1 = test_ref(self.field_master)
        v2 = test_ref(self.field_slave)
        u1 = unknown_ref(self.field_master)
        u2 = unknown_ref(self.field_slave)
        z1 = zero_ref(self.field_master)
        z2 = zero_ref(self.field_slave)
        p = param_ref()

        expr_a = bilin(v1, z2, u1, u2, p)
        expr_b = bilin(z1, v2, u1, u2, p)
        use_backend = self._resolve_backend(None)
        if use_backend == "numpy":
            res_form = compile_mixed_surface_residual_numpy({self.field_master: expr_a, self.field_slave: expr_b})
        else:
            res_form = compile_mixed_surface_residual({self.field_master: expr_a, self.field_slave: expr_b})
        return self.assemble_jacobian(
            res_form,
            {self.field_master: u_master, self.field_slave: u_slave},
            params,
            normal_sign=None,
            normal_source=normal_source,
            sparse=sparse,
            backend=use_backend,
        )


def _field_n_dofs(
    *,
    n_nodes: int,
    n_facets: int,
    value_dim: int,
    space_mode: str,
    facet_dofs: np.ndarray | None,
) -> int:
    if space_mode == "p0":
        if facet_dofs is not None:
            arr = np.asarray(facet_dofs, dtype=int)
            if arr.size == 0:
                return 0
            if np.any(arr < 0):
                raise ValueError("facet_dofs must be non-negative.")
            return int(arr.max()) + 1
        return int(n_facets) * int(value_dim)
    return int(n_nodes) * int(value_dim)


def _contact_space_side_n_dofs(space: "ContactSurfaceSpace", *, side: str) -> int:
    if side == "master":
        return _field_n_dofs(
            n_nodes=int(np.asarray(space.surface_master.coords).shape[0]),
            n_facets=int(np.asarray(space.surface_master.conn).shape[0]),
            value_dim=int(space.value_dim_master),
            space_mode=space.space_mode_master,
            facet_dofs=space.facet_dofs_master,
        )
    if side == "slave":
        return _field_n_dofs(
            n_nodes=int(np.asarray(space.surface_slave.coords).shape[0]),
            n_facets=int(np.asarray(space.surface_slave.conn).shape[0]),
            value_dim=int(space.value_dim_slave),
            space_mode=space.space_mode_slave,
            facet_dofs=space.facet_dofs_slave,
        )
    raise ValueError("side must be 'master' or 'slave'")


@dataclass(eq=False)
class OneToManyContactSurfaceSpace:
    """One-master/multi-slave wrapper built from pairwise ContactSurfaceSpace objects."""

    contacts: tuple[ContactSurfaceSpace, ...]
    field_master: str = "master"
    field_slave: str = "slave"

    @classmethod
    def from_meshes(
        cls,
        master_mesh: BaseMesh,
        slave_meshes: Sequence[BaseMesh],
        *,
        master_facets: np.ndarray | None = None,
        slave_facets_list: Sequence[np.ndarray] | None = None,
        master_facet_selector: Callable[[BaseMesh], np.ndarray] | None = None,
        slave_facet_selectors: Sequence[Callable[[BaseMesh], np.ndarray] | None] | Callable[[BaseMesh], np.ndarray] | None = None,
        master_space: object | None = None,
        slave_spaces: Sequence[object | None] | object | None = None,
        value_dim_master: int | None = None,
        value_dim_slaves: Sequence[int | None] | int | None = None,
        mode_master: str = "touching",
        mode_slave: str = "touching",
        field_master: str = "master",
        field_slave: str = "slave",
        quad_order: int = 1,
        space_mode_master: str = "nodal",
        space_mode_slave: str = "nodal",
        facet_dofs_master: np.ndarray | None = None,
        facet_dofs_slave: np.ndarray | None = None,
        normal_sign: float | None = None,
        tol: float = 1e-8,
        backend: str = "jax",
        fd_eps: float = 1e-6,
        fd_mode: str = "central",
        fd_block_size: int = 1,
        batch_jac: bool | None = None,
        setup_cache_enabled: bool | None = None,
        setup_cache_trace: bool | None = None,
    ) -> "OneToManyContactSurfaceSpace":
        if len(slave_meshes) == 0:
            raise ValueError("slave_meshes must contain at least one mesh.")
        n_slaves = len(slave_meshes)

        if master_facets is None:
            if master_facet_selector is None:
                raise ValueError("Provide either master_facets or master_facet_selector.")
            master_facets = np.asarray(master_facet_selector(master_mesh), dtype=int)
        else:
            master_facets = np.asarray(master_facets, dtype=int)

        if slave_facets_list is None:
            if slave_facet_selectors is None:
                raise ValueError("Provide either slave_facets_list or slave_facet_selectors.")
            if callable(slave_facet_selectors):
                slave_facets_list = [np.asarray(slave_facet_selectors(mesh), dtype=int) for mesh in slave_meshes]
            else:
                if len(slave_facet_selectors) != n_slaves:
                    raise ValueError("slave_facet_selectors length must match slave_meshes length.")
                out_facets: list[np.ndarray] = []
                for mesh, sel in zip(slave_meshes, slave_facet_selectors):
                    if sel is None:
                        raise ValueError("slave_facet_selectors contains None; provide a selector for each slave.")
                    out_facets.append(np.asarray(sel(mesh), dtype=int))
                slave_facets_list = out_facets
        else:
            if len(slave_facets_list) != n_slaves:
                raise ValueError("slave_facets_list length must match slave_meshes length.")
            slave_facets_list = [np.asarray(facets, dtype=int) for facets in slave_facets_list]

        if slave_spaces is None:
            slave_spaces = [None] * n_slaves
        elif isinstance(slave_spaces, Sequence) and not isinstance(slave_spaces, (str, bytes)):
            if len(slave_spaces) != n_slaves:
                raise ValueError("slave_spaces length must match slave_meshes length.")
            slave_spaces = list(slave_spaces)
        else:
            slave_spaces = [slave_spaces] * n_slaves

        if value_dim_slaves is None:
            value_dim_slaves = [None] * n_slaves
        elif isinstance(value_dim_slaves, Sequence) and not isinstance(value_dim_slaves, (str, bytes)):
            if len(value_dim_slaves) != n_slaves:
                raise ValueError("value_dim_slaves length must match slave_meshes length.")
            value_dim_slaves = list(value_dim_slaves)
        else:
            value_dim_slaves = [int(value_dim_slaves)] * n_slaves

        master_side = ContactSide.from_facets(
            master_mesh,
            master_facets,
            master_space,
            value_dim=value_dim_master,
            mode=mode_master,
        )
        slave_sides = [
            ContactSide.from_facets(
                mesh,
                np.asarray(facets, dtype=int),
                space,
                value_dim=value_dim,
                mode=mode_slave,
            )
            for mesh, facets, space, value_dim in zip(slave_meshes, slave_facets_list, slave_spaces, value_dim_slaves)
        ]
        return cls.from_sides(
            master_side,
            slave_sides,
            field_master=field_master,
            field_slave=field_slave,
            quad_order=quad_order,
            space_mode_master=space_mode_master,
            space_mode_slave=space_mode_slave,
            facet_dofs_master=facet_dofs_master,
            facet_dofs_slave=facet_dofs_slave,
            normal_sign=normal_sign,
            tol=tol,
            backend=backend,
            fd_eps=fd_eps,
            fd_mode=fd_mode,
            fd_block_size=fd_block_size,
            batch_jac=batch_jac,
            setup_cache_enabled=setup_cache_enabled,
            setup_cache_trace=setup_cache_trace,
        )

    @classmethod
    def from_sides(
        cls,
        master: ContactSide,
        slaves: Sequence[ContactSide],
        *,
        field_master: str = "master",
        field_slave: str = "slave",
        quad_order: int = 1,
        space_mode_master: str = "nodal",
        space_mode_slave: str = "nodal",
        facet_dofs_master: np.ndarray | None = None,
        facet_dofs_slave: np.ndarray | None = None,
        normal_sign: float | None = None,
        tol: float = 1e-8,
        backend: str = "jax",
        fd_eps: float = 1e-6,
        fd_mode: str = "central",
        fd_block_size: int = 1,
        batch_jac: bool | None = None,
        setup_cache_enabled: bool | None = None,
        setup_cache_trace: bool | None = None,
    ) -> "OneToManyContactSurfaceSpace":
        if len(slaves) == 0:
            raise ValueError("slaves must contain at least one ContactSide.")
        contacts = tuple(
            ContactSurfaceSpace.from_sides(
                master,
                slave,
                field_master=field_master,
                field_slave=field_slave,
                quad_order=quad_order,
                space_mode_master=space_mode_master,
                space_mode_slave=space_mode_slave,
                facet_dofs_master=facet_dofs_master,
                facet_dofs_slave=facet_dofs_slave,
                normal_sign=normal_sign,
                tol=tol,
                backend=backend,
                fd_eps=fd_eps,
                fd_mode=fd_mode,
                fd_block_size=fd_block_size,
                batch_jac=batch_jac,
                setup_cache_enabled=setup_cache_enabled,
                setup_cache_trace=setup_cache_trace,
            )
            for slave in slaves
        )
        return cls(contacts=contacts, field_master=field_master, field_slave=field_slave)

    @classmethod
    def from_surfaces(
        cls,
        surface_master: SurfaceMesh,
        surface_slaves: Sequence[SurfaceMesh],
        *,
        elem_conn_master: np.ndarray | None = None,
        elem_conn_slaves: Sequence[np.ndarray | None] | None = None,
        value_dim_master: int = 1,
        value_dim_slave: int = 1,
        field_master: str = "master",
        field_slave: str = "slave",
        quad_order: int = 1,
        space_mode_master: str = "nodal",
        space_mode_slave: str = "nodal",
        facet_dofs_master: np.ndarray | None = None,
        facet_dofs_slave: np.ndarray | None = None,
        normal_sign: float | None = None,
        tol: float = 1e-8,
        backend: str = "jax",
        fd_eps: float = 1e-6,
        fd_mode: str = "central",
        fd_block_size: int = 1,
        batch_jac: bool | None = None,
        setup_cache_enabled: bool | None = None,
        setup_cache_trace: bool | None = None,
    ) -> "OneToManyContactSurfaceSpace":
        if len(surface_slaves) == 0:
            raise ValueError("surface_slaves must contain at least one surface.")
        if elem_conn_slaves is None:
            elem_conn_slaves = [None] * len(surface_slaves)
        if len(elem_conn_slaves) != len(surface_slaves):
            raise ValueError("elem_conn_slaves length must match surface_slaves length.")
        contacts = tuple(
            ContactSurfaceSpace.from_surfaces(
                surface_master,
                surface_slave,
                elem_conn_master=elem_conn_master,
                elem_conn_slave=elem_conn_slave,
                field_master=field_master,
                field_slave=field_slave,
                value_dim_master=value_dim_master,
                value_dim_slave=value_dim_slave,
                space_mode_master=space_mode_master,
                space_mode_slave=space_mode_slave,
                facet_dofs_master=facet_dofs_master,
                facet_dofs_slave=facet_dofs_slave,
                quad_order=quad_order,
                normal_sign=normal_sign,
                tol=tol,
                backend=backend,
                fd_eps=fd_eps,
                fd_mode=fd_mode,
                fd_block_size=fd_block_size,
                batch_jac=batch_jac,
                setup_cache_enabled=setup_cache_enabled,
                setup_cache_trace=setup_cache_trace,
            )
            for surface_slave, elem_conn_slave in zip(surface_slaves, elem_conn_slaves)
        )
        return cls(contacts=contacts, field_master=field_master, field_slave=field_slave)

    def _split_fields(
        self, u: Mapping[str, npt.ArrayLike] | Sequence[Any]
    ) -> tuple[npt.ArrayLike, list[npt.ArrayLike]]:
        if isinstance(u, Mapping):
            if self.field_master not in u:
                raise KeyError(f"u mapping must contain master field '{self.field_master}'.")
            if "slaves" not in u:
                raise KeyError("u mapping must contain key 'slaves' with per-slave states.")
            u_master = u[self.field_master]
            u_slaves = list(u["slaves"])
        else:
            if len(u) != 2:
                raise ValueError("u must be a mapping or a sequence like (u_master, u_slaves).")
            u_master = u[0]
            u_slaves = list(u[1])
        if len(u_slaves) != len(self.contacts):
            raise ValueError(
                f"u_slaves length mismatch: got {len(u_slaves)}, expected {len(self.contacts)}."
            )
        return u_master, u_slaves

    def _dof_layout(self) -> tuple[int, list[int], int]:
        if len(self.contacts) == 0:
            return 0, [], 0
        n_master = _contact_space_side_n_dofs(self.contacts[0], side="master")
        slave_sizes = [_contact_space_side_n_dofs(contact, side="slave") for contact in self.contacts]
        total = int(n_master + sum(slave_sizes))
        return n_master, slave_sizes, total

    @staticmethod
    def _scatter_pair_indices(local_idx: np.ndarray, *, n_master: int, slave_offset: int) -> np.ndarray:
        idx = np.asarray(local_idx, dtype=int)
        out = np.empty_like(idx)
        master_mask = idx < int(n_master)
        out[master_mask] = idx[master_mask]
        out[~master_mask] = int(n_master) + int(slave_offset) + (idx[~master_mask] - int(n_master))
        return out

    def assemble_residual(
        self,
        res_form: MixedSurfaceResidualForm,
        u: Mapping[str, npt.ArrayLike] | Sequence[Any],
        params: "WeakParams",
        *,
        normal_source: str = "master",
    ) -> np.ndarray:
        u_master, u_slaves = self._split_fields(u)
        n_master, slave_sizes, n_total = self._dof_layout()
        R = np.zeros((n_total,), dtype=float)
        slave_offset = 0
        for contact, u_slave, n_slave in zip(self.contacts, u_slaves, slave_sizes):
            r_local = np.asarray(
                contact.assemble_residual(
                    res_form,
                    {self.field_master: u_master, self.field_slave: u_slave},
                    params,
                    normal_source=normal_source,
                ),
                dtype=float,
            )
            if r_local.shape[0] != n_master + n_slave:
                raise ValueError("Pair residual size mismatch while assembling one-to-many residual.")
            idx = np.concatenate(
                (
                    np.arange(n_master, dtype=int),
                    n_master + slave_offset + np.arange(n_slave, dtype=int),
                )
            )
            R[idx] += r_local
            slave_offset += n_slave
        return R

    def assemble_jacobian(
        self,
        res_form: MixedSurfaceResidualForm,
        u: Mapping[str, npt.ArrayLike] | Sequence[Any],
        params: "WeakParams",
        *,
        normal_source: str = "master",
        sparse: bool = False,
        backend: str | None = None,
        batch_jac: bool | None = None,
    ) -> ContactJacobianReturn:
        u_master, u_slaves = self._split_fields(u)
        n_master, slave_sizes, n_total = self._dof_layout()
        slave_offset = 0
        if sparse:
            rows_all: list[np.ndarray] = []
            cols_all: list[np.ndarray] = []
            data_all: list[np.ndarray] = []
            for contact, u_slave, n_slave in zip(self.contacts, u_slaves, slave_sizes):
                j_local = contact.assemble_jacobian(
                    res_form,
                    {self.field_master: u_master, self.field_slave: u_slave},
                    params,
                    normal_source=normal_source,
                    sparse=True,
                    backend=backend,
                    batch_jac=batch_jac,
                )
                rows, cols, data, n_pair = j_local
                if int(n_pair) != n_master + n_slave:
                    raise ValueError("Pair Jacobian size mismatch while assembling sparse one-to-many Jacobian.")
                rows_all.append(
                    self._scatter_pair_indices(np.asarray(rows, dtype=int), n_master=n_master, slave_offset=slave_offset)
                )
                cols_all.append(
                    self._scatter_pair_indices(np.asarray(cols, dtype=int), n_master=n_master, slave_offset=slave_offset)
                )
                data_all.append(np.asarray(data, dtype=float))
                slave_offset += n_slave
            if rows_all:
                rows_out = np.concatenate(rows_all)
                cols_out = np.concatenate(cols_all)
                data_out = np.concatenate(data_all)
            else:
                rows_out = np.zeros((0,), dtype=int)
                cols_out = np.zeros((0,), dtype=int)
                data_out = np.zeros((0,), dtype=float)
            return rows_out, cols_out, data_out, n_total

        K = np.zeros((n_total, n_total), dtype=float)
        for contact, u_slave, n_slave in zip(self.contacts, u_slaves, slave_sizes):
            j_local = np.asarray(
                contact.assemble_jacobian(
                    res_form,
                    {self.field_master: u_master, self.field_slave: u_slave},
                    params,
                    normal_source=normal_source,
                    sparse=False,
                    backend=backend,
                    batch_jac=batch_jac,
                ),
                dtype=float,
            )
            if j_local.shape != (n_master + n_slave, n_master + n_slave):
                raise ValueError("Pair Jacobian shape mismatch while assembling dense one-to-many Jacobian.")
            idx = np.concatenate(
                (
                    np.arange(n_master, dtype=int),
                    n_master + slave_offset + np.arange(n_slave, dtype=int),
                )
            )
            K[np.ix_(idx, idx)] += j_local
            slave_offset += n_slave
        return K

    def assemble_bilinear(
        self,
        bilin: Callable[..., Any],
        u_master: Mapping[str, npt.ArrayLike] | Sequence[Any] | npt.ArrayLike,
        u_slaves: Sequence[npt.ArrayLike] | None = None,
        params: "WeakParams" | None = None,
        *,
        sparse: bool = False,
        normal_source: str = "master",
    ) -> ContactJacobianReturn:
        if params is None:
            if u_slaves is None:
                raise TypeError("params is required")
            if isinstance(u_master, Mapping) or (isinstance(u_master, Sequence) and not hasattr(u_master, "shape")):
                params = u_slaves  # type: ignore[assignment]
                u_master, u_slaves = self._split_fields(u_master)  # type: ignore[arg-type]
            else:
                raise TypeError("params is required")
        elif u_slaves is None:
            u_master, u_slaves = self._split_fields(u_master)  # type: ignore[arg-type]
        assert params is not None
        assert u_slaves is not None

        n_master, slave_sizes, n_total = self._dof_layout()
        slave_offset = 0
        if sparse:
            rows_all: list[np.ndarray] = []
            cols_all: list[np.ndarray] = []
            data_all: list[np.ndarray] = []
            for contact, u_slave, n_slave in zip(self.contacts, u_slaves, slave_sizes):
                j_local = contact.assemble_bilinear(
                    bilin,
                    u_master,
                    u_slave,
                    params,
                    sparse=True,
                    normal_source=normal_source,
                )
                rows, cols, data, n_pair = j_local
                if int(n_pair) != n_master + n_slave:
                    raise ValueError("Pair Jacobian size mismatch while assembling sparse one-to-many bilinear.")
                rows_all.append(
                    self._scatter_pair_indices(np.asarray(rows, dtype=int), n_master=n_master, slave_offset=slave_offset)
                )
                cols_all.append(
                    self._scatter_pair_indices(np.asarray(cols, dtype=int), n_master=n_master, slave_offset=slave_offset)
                )
                data_all.append(np.asarray(data, dtype=float))
                slave_offset += n_slave
            if rows_all:
                rows_out = np.concatenate(rows_all)
                cols_out = np.concatenate(cols_all)
                data_out = np.concatenate(data_all)
            else:
                rows_out = np.zeros((0,), dtype=int)
                cols_out = np.zeros((0,), dtype=int)
                data_out = np.zeros((0,), dtype=float)
            return rows_out, cols_out, data_out, n_total

        K = np.zeros((n_total, n_total), dtype=float)
        for contact, u_slave, n_slave in zip(self.contacts, u_slaves, slave_sizes):
            j_local = np.asarray(
                contact.assemble_bilinear(
                    bilin,
                    u_master,
                    u_slave,
                    params,
                    sparse=False,
                    normal_source=normal_source,
                ),
                dtype=float,
            )
            if j_local.shape != (n_master + n_slave, n_master + n_slave):
                raise ValueError("Pair Jacobian shape mismatch while assembling dense one-to-many bilinear.")
            idx = np.concatenate(
                (
                    np.arange(n_master, dtype=int),
                    n_master + slave_offset + np.arange(n_slave, dtype=int),
                )
            )
            K[np.ix_(idx, idx)] += j_local
            slave_offset += n_slave
        return K

    def assemble_contact_coupling_matrices(self):
        from .contact_interface import ContactCouplingMatrix

        n_master, slave_sizes, _ = self._dof_layout()
        n_slaves_total = int(sum(slave_sizes))
        rows_mm: list[np.ndarray] = []
        cols_mm: list[np.ndarray] = []
        data_mm: list[np.ndarray] = []
        rows_ms: list[np.ndarray] = []
        cols_ms: list[np.ndarray] = []
        data_ms: list[np.ndarray] = []

        slave_offset = 0
        for contact, n_slave in zip(self.contacts, slave_sizes):
            m_mm, m_ms_local = contact.assemble_contact_coupling_matrices()
            rows_mm.append(np.asarray(m_mm.rows, dtype=int))
            cols_mm.append(np.asarray(m_mm.cols, dtype=int))
            data_mm.append(np.asarray(m_mm.data, dtype=float))
            rows_ms.append(np.asarray(m_ms_local.rows, dtype=int))
            cols_ms.append(np.asarray(m_ms_local.cols, dtype=int) + slave_offset)
            data_ms.append(np.asarray(m_ms_local.data, dtype=float))
            if m_mm.shape != (n_master, n_master):
                raise ValueError("Pair M_aa shape mismatch while assembling one-to-many coupling matrices.")
            if m_ms_local.shape != (n_master, n_slave):
                raise ValueError("Pair M_ab shape mismatch while assembling one-to-many coupling matrices.")
            slave_offset += n_slave

        mm = ContactCouplingMatrix(
            rows=np.concatenate(rows_mm) if rows_mm else np.zeros((0,), dtype=int),
            cols=np.concatenate(cols_mm) if cols_mm else np.zeros((0,), dtype=int),
            data=np.concatenate(data_mm) if data_mm else np.zeros((0,), dtype=float),
            shape=(n_master, n_master),
        )
        ms = ContactCouplingMatrix(
            rows=np.concatenate(rows_ms) if rows_ms else np.zeros((0,), dtype=int),
            cols=np.concatenate(cols_ms) if cols_ms else np.zeros((0,), dtype=int),
            data=np.concatenate(data_ms) if data_ms else np.zeros((0,), dtype=float),
            shape=(n_master, n_slaves_total),
        )
        return mm, ms

    def assemble_contact_kkt(
        self,
        *,
        rho: float = 0.0,
        multiplier_space: str = "nodal",
        backend: str = "numpy",
        format: str = "fluxsparse",
        return_blocks: bool = False,
    ):
        m_aa, m_ab = self.assemble_contact_coupling_matrices()
        master_facets = np.asarray(self.contacts[0].surface_master.conn, dtype=int)
        return assemble_contact_kkt(
            m_aa,
            m_ab,
            rho=rho,
            multiplier_space=multiplier_space,
            facet_conn_master=master_facets,
            backend=backend,
            format=format,
            return_blocks=return_blocks,
        )

    def assemble_contact_operators(
        self,
        *,
        method: str = "mortar",
        rho: float = 0.0,
        multiplier_space: str = "nodal",
        backend: str = "numpy",
        res_form: MixedSurfaceResidualForm | None = None,
        u: Mapping[str, npt.ArrayLike] | Sequence[Any] | None = None,
        params: "WeakParams" | None = None,
        normal_source: str = "master",
        sparse: bool = False,
        batch_jac: bool | None = None,
    ) -> ContactOperators:
        return assemble_contact_operators(
            self,
            method=method,
            rho=rho,
            multiplier_space=multiplier_space,
            backend=backend,
            res_form=res_form,
            u=u,
            params=params,
            normal_source=normal_source,
            sparse=sparse,
            batch_jac=batch_jac,
        )


__all__ = [
    "ContactSide",
    "OneSidedContact",
    "OneSidedContactSurfaceSpace",
    "ContactSurfaceSpace",
    "OneToManyContactSurfaceSpace",
    "ContactOperators",
    "assemble_contact_operators",
    "assemble_contact_interface_residual",
    "assemble_contact_interface_jacobian",
    "assemble_contact_coupling_matrices",
    "assemble_contact_kkt",
    "solve_contact_kkt",
    "facet_gap_values",
    "active_contact_facets",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Optional, Protocol, Sequence, TYPE_CHECKING, TypeVar, cast
import jax
import jax.numpy as jnp

from .dtypes import INDEX_DTYPE
import numpy as np
import numpy.typing as npt

DTYPE = jnp.float64 if jax.config.read("jax_enable_x64") else jnp.float32

from .base import BaseMesh, BaseMeshPytree
from .hex import HexMesh, HexMeshPytree

P = TypeVar("P")

if TYPE_CHECKING:
    from ..solver.bc import SurfaceFormContext


class SurfaceSpaceLike(Protocol):
    value_dim: int
    mesh: BaseMesh


SurfaceLinearForm = Callable[["SurfaceFormContext", P], npt.ArrayLike]


def _polygon_area(pts: np.ndarray) -> float:
    """
    Polygon area in 3D by fan triangulation (works for tri/quad faces).
    Assumes points are planar and ordered.
    """
    if pts.shape[0] < 3:
        return 0.0
    area = 0.0
    p0 = pts[0]
    for i in range(1, pts.shape[0] - 1):
        v1 = pts[i] - p0
        v2 = pts[i + 1] - p0
        area += float(0.5 * np.linalg.norm(np.cross(v1, v2)))
    return float(area)


@dataclass(eq=False)
class SurfaceMesh(BaseMesh):
    """
    Simple boundary mesh made of facets (tri/quad) that live in the volume mesh nodes.
    Uses BaseMesh.conn to store facets.
    """

    facet_tags: Optional[jnp.ndarray] = None

    def __post_init__(self):
        # Keep facet_tags mirrored in cell_tags for BaseMesh compat.
        if self.cell_tags is None and self.facet_tags is not None:
            self.cell_tags = self.facet_tags
        if self.facet_tags is None and self.cell_tags is not None:
            self.facet_tags = self.cell_tags

    @classmethod
    def from_facets(
        cls,
        coords: jnp.ndarray,
        facets: jnp.ndarray,
        facet_tags: Optional[jnp.ndarray] = None,
        node_tags: Optional[jnp.ndarray] = None,
    ) -> "SurfaceMesh":
        coords_j = jnp.asarray(coords, dtype=DTYPE)
        facets_j = jnp.asarray(facets, dtype=INDEX_DTYPE)
        tags_j = None if facet_tags is None else jnp.asarray(facet_tags, dtype=INDEX_DTYPE)
        node_tags_j = None if node_tags is None else jnp.asarray(node_tags)
        return cls(coords=coords_j, conn=facets_j, cell_tags=tags_j, node_tags=node_tags_j, facet_tags=tags_j)

    @classmethod
    def from_hex_mesh(
        cls,
        mesh: HexMesh,
        facets: jnp.ndarray,
        facet_tags: Optional[jnp.ndarray] = None,
    ) -> "SurfaceMesh":
        """
        Build a surface mesh that reuses the volume mesh coordinates.
        Facets must reference the volume mesh node numbering.
        """
        return cls.from_facets(mesh.coords, facets, facet_tags=facet_tags, node_tags=mesh.node_tags)

    @property
    def n_facets(self) -> int:
        return self.n_elems

    def facet_areas(self) -> np.ndarray:
        """Return per-facet area (uses NumPy for simplicity)."""
        coords = np.asarray(self.coords)
        facets = np.asarray(self.conn, dtype=int)
        areas = np.zeros(facets.shape[0], dtype=float)
        for i, nodes in enumerate(facets):
            pts = coords[nodes]
            areas[i] = _polygon_area(pts)
        return areas

    def select_by_tag(self, tag: int) -> "SurfaceMesh":
        """Return a new SurfaceMesh containing only facets with given tag."""
        if self.facet_tags is None:
            raise ValueError("facet_tags not set on this SurfaceMesh")
        mask = np.asarray(self.facet_tags) == tag
        return SurfaceMesh.from_facets(
            self.coords,
            self.conn[mask],
            facet_tags=self.facet_tags[mask],
            node_tags=self.node_tags,
        )

    def facet_normals(self, *, outward_from=None, normalize: bool = True) -> np.ndarray:
        from ..solver.bc import facet_normals
        return facet_normals(self, outward_from=outward_from, normalize=normalize)

    def assemble_load(
        self,
        load: npt.ArrayLike,
        *,
        dim: int,
        n_total_nodes: int | None = None,
        F0: npt.ArrayLike | None = None,
    ) -> np.ndarray:
        from ..solver.bc import assemble_surface_load
        return assemble_surface_load(self, load, dim=dim, n_total_nodes=n_total_nodes, F0=F0)

    def assemble_linear_form(
        self,
        form: SurfaceLinearForm[P],
        params: P,
        *,
        dim: int,
        n_total_nodes: int | None = None,
        F0: npt.ArrayLike | None = None,
    ) -> np.ndarray:
        from ..solver.bc import assemble_surface_linear_form
        return assemble_surface_linear_form(self, form, params, dim=dim, n_total_nodes=n_total_nodes, F0=F0)

    def assemble_linear_form_on_space(
        self,
        space: SurfaceSpaceLike,
        form: SurfaceLinearForm[P],
        params: P,
        *,
        F0: npt.ArrayLike | None = None,
    ) -> np.ndarray:
        """
        Assemble surface linear form using global size inferred from a volume space.
        """
        dim = int(getattr(space, "value_dim", 1))
        mesh = cast(BaseMesh, getattr(space, "mesh", self))
        n_total_nodes = int(mesh.n_nodes)
        return self.assemble_linear_form(form, params, dim=dim, n_total_nodes=n_total_nodes, F0=F0)

    def assemble_bilinear_form_on_space(
        self,
        space,
        form,
        params,
        *,
        pattern=None,
        return_flux_matrix: bool = True,
    ):
        """
        Assemble surface bilinear form using global size inferred from a volume space.
        """
        from ..solver.bc import assemble_surface_bilinear_form
        dim = int(getattr(space, "value_dim", 1))
        n_total_nodes = int(getattr(space, "mesh", self).n_nodes)
        if pattern is None and hasattr(space, "get_sparsity_pattern"):
            pattern = space.get_sparsity_pattern(with_idx=True)
        return assemble_surface_bilinear_form(
            self,
            form,
            params,
            dim=dim,
            n_total_nodes=n_total_nodes,
            pattern=pattern,
            return_flux_matrix=return_flux_matrix,
        )


@dataclass(frozen=True)
class SurfaceWithElemConn:
    surface: SurfaceMesh
    elem_conn: np.ndarray


def surface_with_elem_conn(mesh: BaseMesh, facets, *, mode: str = "touching") -> SurfaceWithElemConn:
    """
    Build a SurfaceMesh from facets and return it with a matching elem_conn array.
    """
    surface = SurfaceMesh.from_facets(mesh.coords, facets, node_tags=mesh.node_tags)
    elems = mesh.elements_from_facets(facets, mode=mode)
    elem_conn = np.asarray(mesh.conn, dtype=int)[elems]
    return SurfaceWithElemConn(surface=surface, elem_conn=elem_conn)

    def assemble_traction(
        self,
        traction: float | Sequence[float],
        *,
        dim: int = 3,
        n_total_nodes: int | None = None,
        F0: npt.ArrayLike | None = None,
        outward_from: npt.ArrayLike | None = None,
    ) -> np.ndarray:
        from ..solver.bc import assemble_surface_traction
        return assemble_surface_traction(
            self,
            traction,
            dim=dim,
            n_total_nodes=n_total_nodes,
            F0=F0,
            outward_from=outward_from,
        )


@jax.tree_util.register_pytree_node_class
@dataclass(eq=False)
class SurfaceMeshPytree(BaseMeshPytree):
    """
    Simple boundary mesh made of facets (tri/quad) that live in the volume mesh nodes.
    Uses BaseMesh.conn to store facets.
    """

    facet_tags: Optional[jnp.ndarray] = None

    def __post_init__(self):
        if self.cell_tags is None and self.facet_tags is not None:
            self.cell_tags = self.facet_tags
        if self.facet_tags is None and self.cell_tags is not None:
            self.facet_tags = self.cell_tags

    @classmethod
    def from_facets(
        cls,
        coords: jnp.ndarray,
        facets: jnp.ndarray,
        facet_tags: Optional[jnp.ndarray] = None,
        node_tags: Optional[jnp.ndarray] = None,
    ) -> "SurfaceMeshPytree":
        coords_j = jnp.asarray(coords, dtype=DTYPE)
        facets_j = jnp.asarray(facets, dtype=INDEX_DTYPE)
        tags_j = None if facet_tags is None else jnp.asarray(facet_tags, dtype=INDEX_DTYPE)
        node_tags_j = None if node_tags is None else jnp.asarray(node_tags)
        return cls(coords=coords_j, conn=facets_j, cell_tags=tags_j, node_tags=node_tags_j, facet_tags=tags_j)

    @classmethod
    def from_hex_mesh(
        cls,
        mesh: HexMesh | HexMeshPytree,
        facets: jnp.ndarray,
        facet_tags: Optional[jnp.ndarray] = None,
    ) -> "SurfaceMeshPytree":
        return cls.from_facets(mesh.coords, facets, facet_tags=facet_tags, node_tags=mesh.node_tags)

    @property
    def n_facets(self) -> int:
        return self.n_elems

    def facet_areas(self) -> np.ndarray:
        coords = np.asarray(self.coords)
        facets = np.asarray(self.conn, dtype=int)
        areas = np.zeros(facets.shape[0], dtype=float)
        for i, nodes in enumerate(facets):
            pts = coords[nodes]
            areas[i] = _polygon_area(pts)
        return areas

    def select_by_tag(self, tag: int) -> "SurfaceMeshPytree":
        if self.facet_tags is None:
            raise ValueError("facet_tags not set on this SurfaceMesh")
        mask = np.asarray(self.facet_tags) == tag
        return SurfaceMeshPytree.from_facets(
            self.coords,
            self.conn[mask],
            facet_tags=self.facet_tags[mask],
            node_tags=self.node_tags,
        )

    def facet_normals(self, *, outward_from=None, normalize: bool = True) -> np.ndarray:
        from ..solver.bc import facet_normals
        return facet_normals(self, outward_from=outward_from, normalize=normalize)

    def assemble_load(
        self,
        load: npt.ArrayLike,
        *,
        dim: int,
        n_total_nodes: int | None = None,
        F0: npt.ArrayLike | None = None,
    ) -> np.ndarray:
        from ..solver.bc import assemble_surface_load
        return assemble_surface_load(self, load, dim=dim, n_total_nodes=n_total_nodes, F0=F0)

    def assemble_linear_form(
        self,
        form: SurfaceLinearForm[P],
        params: P,
        *,
        dim: int,
        n_total_nodes: int | None = None,
        F0: npt.ArrayLike | None = None,
    ) -> np.ndarray:
        from ..solver.bc import assemble_surface_linear_form
        return assemble_surface_linear_form(self, form, params, dim=dim, n_total_nodes=n_total_nodes, F0=F0)

    def assemble_linear_form_on_space(
        self,
        space: SurfaceSpaceLike,
        form: SurfaceLinearForm[P],
        params: P,
        *,
        F0: npt.ArrayLike | None = None,
    ) -> np.ndarray:
        """
        Assemble surface linear form using global size inferred from a volume space.
        """
        dim = int(getattr(space, "value_dim", 1))
        mesh = cast(BaseMesh, getattr(space, "mesh", self))
        n_total_nodes = int(mesh.n_nodes)
        return self.assemble_linear_form(form, params, dim=dim, n_total_nodes=n_total_nodes, F0=F0)

    def assemble_bilinear_form_on_space(
        self,
        space,
        form,
        params,
        *,
        pattern=None,
        return_flux_matrix: bool = True,
    ):
        """
        Assemble surface bilinear form using global size inferred from a volume space.
        """
        from ..solver.bc import assemble_surface_bilinear_form
        dim = int(getattr(space, "value_dim", 1))
        n_total_nodes = int(getattr(space, "mesh", self).n_nodes)
        if pattern is None and hasattr(space, "get_sparsity_pattern"):
            pattern = space.get_sparsity_pattern(with_idx=True)
        return assemble_surface_bilinear_form(
            self,
            form,
            params,
            dim=dim,
            n_total_nodes=n_total_nodes,
            pattern=pattern,
            return_flux_matrix=return_flux_matrix,
        )

    def assemble_traction(
        self,
        traction: float | Sequence[float],
        *,
        dim: int = 3,
        n_total_nodes: int | None = None,
        F0: npt.ArrayLike | None = None,
        outward_from: npt.ArrayLike | None = None,
    ) -> np.ndarray:
        from ..solver.bc import assemble_surface_traction
        return assemble_surface_traction(
            self,
            traction,
            dim=dim,
            n_total_nodes=n_total_nodes,
            F0=F0,
            outward_from=outward_from,
        )

    def assemble_flux(
        self,
        flux: npt.ArrayLike,
        *,
        n_total_nodes: int | None = None,
        F0: npt.ArrayLike | None = None,
        outward_from: npt.ArrayLike | None = None,
    ) -> np.ndarray:
        from ..solver.bc import assemble_surface_flux
        return assemble_surface_flux(
            self,
            flux,
            n_total_nodes=n_total_nodes,
            F0=F0,
            outward_from=outward_from,
        )



from __future__ import annotations

from typing import Any, Optional

import jax
import jax.numpy as jnp

from .dtypes import INDEX_DTYPE
from .forms import FormContext


def assemble_jacobian_global(
    space,
    res_form,
    u: jnp.ndarray,
    params: Any,
):
    """
    Assemble Jacobian (dR/du) from element residual res_form.
    res_form(ctx, u_elem, params) -> (n_q, n_ldofs)
    """
    from . import assembly as _a
    from ..solver import FluxSparseMatrix

    elem_dofs = space.elem_dofs
    n_dofs = space.n_dofs
    n_ldofs = space.n_ldofs

    elem_data = space.build_form_contexts()

    def fe_fun(u_elem, ctx: FormContext, elem_id):
        ctx_with_id = FormContext(ctx.test, ctx.trial, ctx.x_q, ctx.w, elem_id)
        integrand = res_form(ctx_with_id, u_elem, params)
        wJ = ctx.w * ctx.test.detJ
        fe = _a._integrate_q_linear(
            integrand,
            wJ,
            includes_measure=bool(getattr(res_form, "_includes_measure", False)),
        )
        return fe

    jac_fun = jax.jacrev(fe_fun, argnums=0)

    u_elems = u[elem_dofs]
    elem_ids = jnp.arange(elem_dofs.shape[0], dtype=INDEX_DTYPE)
    J_e_all = jax.vmap(jac_fun)(u_elems, elem_data, elem_ids)

    pat = _a._get_pattern(space, with_idx=False)
    if pat is None:
        rows = jnp.repeat(elem_dofs, n_ldofs, axis=1).reshape(-1)
        cols = jnp.tile(elem_dofs, (1, n_ldofs)).reshape(-1)
    else:
        rows = pat.rows
        cols = pat.cols
    data = J_e_all.reshape(-1)
    return FluxSparseMatrix(rows, cols, data, n_dofs)


def assemble_jacobian_elementwise(
    space,
    res_form,
    u: jnp.ndarray,
    params: Any,
):
    """
    Assemble Jacobian with element kernels via vmap + scatter_add.
    Recompiles if n_dofs changes, but independent of element count.
    """
    from . import assembly as _a
    from ..solver import FluxSparseMatrix

    elem_dofs = space.elem_dofs
    n_dofs = space.n_dofs
    n_ldofs = space.n_ldofs

    ctxs = space.build_form_contexts()

    def fe_fun(u_elem, ctx: FormContext):
        integrand = res_form(ctx, u_elem, params)
        wJ = ctx.w * ctx.test.detJ
        return _a._integrate_q_linear(
            integrand,
            wJ,
            includes_measure=bool(getattr(res_form, "_includes_measure", False)),
        )

    jac_fun = jax.jacrev(fe_fun, argnums=0)
    u_elems = u[elem_dofs]
    J_e_all = jax.vmap(jac_fun)(u_elems, ctxs)

    pat = _a._get_pattern(space, with_idx=False)
    if pat is None:
        rows = jnp.repeat(elem_dofs, n_ldofs, axis=1).reshape(-1)
        cols = jnp.tile(elem_dofs, (1, n_ldofs)).reshape(-1)
    else:
        rows = pat.rows
        cols = pat.cols
    data = J_e_all.reshape(-1)
    return FluxSparseMatrix(rows, cols, data, n_dofs)


def assemble_jacobian_values(
    space,
    res_form,
    u: jnp.ndarray,
    params: Any,
    *,
    kernel=None,
    n_chunks: Optional[int] = None,
    pad_trace: bool | None = None,
    policy=None,
):
    """
    Assemble only the numeric values for the Jacobian (pattern-free).
    """
    from . import assembly as _a

    include_x_q_req: bool | None = None if policy is not None else False
    n_chunks, include_x_q, lightweight_context, chunk_build_context, pad_trace = _a._resolve_assembly_policy(
        policy=policy,
        n_chunks=n_chunks,
        include_x_q=include_x_q_req,
        lightweight_context=None,
        chunk_build_context=None,
        pad_trace=pad_trace,
    )
    ker = kernel if kernel is not None else _a.make_element_jacobian_kernel(res_form, params)
    vmapped_kernel = jax.vmap(ker)

    u_elems = u[space.elem_dofs]
    if n_chunks is None:
        def _eval(include_x_q_eff: bool):
            ctxs = space.build_form_contexts(include_x_q=include_x_q_eff, lightweight=lightweight_context)
            return vmapped_kernel(u_elems, ctxs)

        try:
            J_e_all = _eval(include_x_q)
        except Exception:
            if include_x_q:
                raise
            J_e_all = _eval(True)
        return J_e_all.reshape(-1)

    n_elems = int(u_elems.shape[0])
    n_chunks, chunk_size, pad, n_pad, valid_mask = _a._prepare_chunk_iteration(
        n_elems=int(n_elems),
        n_chunks=n_chunks,
        pad_trace=pad_trace,
    )
    m = int(space.n_ldofs)
    if pad:
        u_elems_pad = jnp.concatenate([u_elems, jnp.repeat(u_elems[-1:], pad, axis=0)], axis=0)
    else:
        u_elems_pad = u_elems

    def _init_chunk(include_x_q_eff: bool):
        use_chunk_context, conn_pad, elem_ids, ctxs_pad = _a._prepare_chunk_context_source(
            space,
            n_pad=int(n_pad),
            pad=int(pad),
            dep=None,
            include_x_q=include_x_q_eff,
            lightweight_context=lightweight_context,
            chunk_build_context=chunk_build_context,
        )

        def chunk_values_fn(start: int):
            ctx_chunk = _a._chunk_context_from_source(
                space,
                start=start,
                chunk_size=chunk_size,
                use_chunk_context=use_chunk_context,
                conn_pad=conn_pad,
                elem_ids=elem_ids,
                ctxs_pad=ctxs_pad,
                include_x_q=include_x_q_eff,
                lightweight_context=lightweight_context,
            )
            u_chunk = _a._slice_first_dim(u_elems_pad, start, chunk_size)
            return vmapped_kernel(u_chunk, ctx_chunk)

        sample_J = chunk_values_fn(0)[0]
        return sample_J, chunk_values_fn

    try:
        sample_J, chunk_values_fn = _init_chunk(include_x_q)
    except Exception:
        if include_x_q:
            raise
        sample_J, chunk_values_fn = _init_chunk(True)

    data = _a._accumulate_chunk_matrix_data(
        n_chunks=n_chunks,
        chunk_size=chunk_size,
        n_pad=n_pad,
        m=m,
        dtype=sample_J.dtype,
        valid_mask=valid_mask,
        chunk_values_fn=chunk_values_fn,
    )
    return data[: n_elems * m * m]


def assemble_jacobian_scatter(
    space,
    res_form,
    u: jnp.ndarray,
    params: Any,
    *,
    kernel=None,
    pattern=None,
    n_chunks: Optional[int] = None,
    pad_trace: bool | None = None,
    policy=None,
):
    """
    Assemble Jacobian using jitted element kernel + vmap + scatter_add.
    If a SparsityPattern is provided, rows/cols are reused without regeneration.
    """
    from . import assembly as _a
    from ..solver import FluxSparseMatrix

    if pattern is not None:
        pat = pattern
    else:
        pat = _a._get_pattern(space, with_idx=True)
        if pat is None:
            pat = _a.make_sparsity_pattern(space, with_idx=True)
    data = assemble_jacobian_values(
        space,
        res_form,
        u,
        params,
        kernel=kernel,
        n_chunks=n_chunks,
        pad_trace=pad_trace,
        policy=policy,
    )
    return FluxSparseMatrix(pat, data)


def assemble_jacobian(
    space,
    res_form,
    u: jnp.ndarray,
    params: Any,
    *,
    kernel=None,
    pattern=None,
    n_chunks: Optional[int] = None,
    pad_trace: bool | None = None,
    policy=None,
):
    """
    Assemble the global Jacobian (scatter-based).
    If kernel is provided: kernel(u_elem, ctx) -> (n_ldofs, n_ldofs).
    """
    return assemble_jacobian_scatter(
        space,
        res_form,
        u,
        params,
        kernel=kernel,
        pattern=pattern,
        n_chunks=n_chunks,
        pad_trace=pad_trace,
        policy=policy,
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, TYPE_CHECKING, TypeAlias

import numpy as np
import jax.numpy as jnp

try:
    import scipy.sparse as sp
except Exception:  # pragma: no cover
    sp = None

from .sparse import FluxSparseMatrix, coalesce_coo

if TYPE_CHECKING:
    from jax import Array as JaxArray

    ArrayLike: TypeAlias = np.ndarray | JaxArray
else:
    ArrayLike: TypeAlias = np.ndarray
DirichletLike: TypeAlias = tuple[np.ndarray, np.ndarray]


def _normalize_dirichlet_values(dofs: ArrayLike, vals: ArrayLike | None) -> np.ndarray:
    if vals is None:
        return np.zeros(np.asarray(dofs).shape[0], dtype=float)
    arr = np.asarray(vals)
    if arr.ndim == 0:
        return np.full(np.asarray(dofs).shape[0], float(arr), dtype=float)
    return arr


def _normalize_dirichlet(dofs: ArrayLike, vals: ArrayLike | None) -> DirichletLike:
    dir_arr = np.asarray(dofs, dtype=int)
    return dir_arr, _normalize_dirichlet_values(dir_arr, vals)


@dataclass(frozen=True)
class CondensedSystem:
    K: Any
    F: Any
    free_dofs: np.ndarray
    dir_dofs: np.ndarray
    dir_vals: np.ndarray
    n_dofs: int

    def expand(self, u_free: ArrayLike, *, fill_dirichlet: bool = True) -> np.ndarray:
        u_full = np.zeros(self.n_dofs, dtype=np.asarray(u_free).dtype)
        u_full[self.free_dofs] = np.asarray(u_free)
        if fill_dirichlet and self.dir_dofs.size:
            u_full[self.dir_dofs] = np.asarray(self.dir_vals, dtype=u_full.dtype)
        return u_full




@dataclass(frozen=True)
class DirichletBC:
    """
    Dirichlet boundary condition container with helper methods.
    """
    dofs: np.ndarray
    vals: np.ndarray

    def __post_init__(self):
        dofs, vals = _normalize_dirichlet(self.dofs, self.vals)
        object.__setattr__(self, "dofs", dofs)
        object.__setattr__(self, "vals", vals)

    @classmethod
    def from_boundary_dofs(cls, mesh, predicate, *, values: ArrayLike | None = None, **kwargs) -> "DirichletBC":
        """
        Build from mesh.boundary_dofs_where predicate.

        kwargs are forwarded to mesh.boundary_dofs_where (e.g. components=..., dof_per_node=...).
        """
        dofs = mesh.boundary_dofs_where(predicate, **kwargs)
        vals = _normalize_dirichlet_values(dofs, values)
        return cls(dofs, vals)

    @classmethod
    def from_bbox(
        cls,
        mesh,
        *,
        mins: ArrayLike | None = None,
        maxs: ArrayLike | None = None,
        tol: float = 1e-8,
        values: ArrayLike | None = None,
        **kwargs,
    ) -> "DirichletBC":
        """
        Build from the mesh axis-aligned bounding box.

        mins/maxs default to mesh coordinate extrema. kwargs are forwarded to
        mesh.boundary_dofs_where (e.g. components=..., dof_per_node=...).
        """
        from ..mesh.predicate import bbox_predicate

        coords = np.asarray(mesh.coords)
        if mins is None:
            mins = coords.min(axis=0)
        if maxs is None:
            maxs = coords.max(axis=0)
        pred = bbox_predicate(mins, maxs, tol=tol)
        dofs = mesh.boundary_dofs_where(pred, **kwargs)
        vals = _normalize_dirichlet_values(dofs, values)
        return cls(dofs, vals)

    def as_tuple(self) -> tuple[np.ndarray, np.ndarray]:
        return self.dofs, self.vals

    def condense_system(self, A: Any, F: ArrayLike, *, check: bool = True) -> CondensedSystem:
        return condense_dirichlet_system(A, F, self.dofs, self.vals, check=check)

    def enforce_system(self, A: Any, F: ArrayLike):
        return enforce_dirichlet_system(A, F, self.dofs, self.vals)

    def condense_flux(self, A: FluxSparseMatrix, F: ArrayLike):
        """
        Condense for FluxSparseMatrix and return (K_free, F_free, free_dofs).
        """
        condensed = self.condense_system(A, F)
        free = condensed.free_dofs
        return restrict_flux_to_free(A, free), condensed.F, free

    def enforce_flux(self, A: FluxSparseMatrix, F: ArrayLike):
        return enforce_dirichlet_fluxsparse(A, F, self.dofs, self.vals)

    def split_matrix(self, A: Any, *, n_total: int | None = None):
        return split_dirichlet_matrix(A, self.dofs, n_total=n_total)

    def free_dofs(self, n_dofs: int) -> np.ndarray:
        return free_dofs(n_dofs, self.dofs)

    def expand_solution(
        self,
        u_free: ArrayLike,
        *,
        free: np.ndarray | None = None,
        n_total: int | None = None,
    ) -> np.ndarray:
        if free is None:
            if n_total is None:
                raise ValueError("n_total is required when free is not provided.")
            free = free_dofs(n_total, self.dofs)
        if n_total is None:
            max_free = int(np.max(free)) if len(free) else -1
            max_dir = int(np.max(self.dofs)) if len(self.dofs) else -1
            n_total = max(max_free, max_dir) + 1
        return expand_dirichlet_solution(u_free, free, self.dofs, self.vals, n_total=n_total)


def enforce_dirichlet_dense(K, F, dofs, vals):
    """Apply Dirichlet conditions directly to stiffness/load (dense)."""
    Kc = np.asarray(K, dtype=float).copy()
    Fc = np.asarray(F, dtype=float).copy()
    dofs, vals = _normalize_dirichlet(dofs, vals)
    if Fc.ndim == 2:
        Fc = Fc - (Kc[:, dofs] @ vals)[:, None]
    else:
        Fc = Fc - Kc[:, dofs] @ vals
    for d, v in zip(dofs, vals):
        Kc[d, :] = 0.0
        Kc[:, d] = 0.0
        Kc[d, d] = 1.0
        if Fc.ndim == 2:
            Fc[d, :] = v
        else:
            Fc[d] = v
    return Kc, Fc


def enforce_dirichlet_dense_jax(K, F, dofs, vals):
    """Apply Dirichlet conditions directly to stiffness/load (dense, JAX-friendly)."""
    import jax.numpy as jnp

    dofs, vals = _normalize_dirichlet(dofs, vals)
    dofs = jnp.asarray(dofs, dtype=jnp.int32)
    vals = jnp.asarray(vals, dtype=F.dtype)
    if F.ndim == 2:
        F_mod = F - (K[:, dofs] @ vals)[:, None]
    else:
        F_mod = F - K[:, dofs] @ vals
    K_mod = K.at[:, dofs].set(0.0)
    K_mod = K_mod.at[dofs, :].set(0.0)
    K_mod = K_mod.at[dofs, dofs].set(1.0)
    if F.ndim == 2:
        F_mod = F_mod.at[dofs, :].set(vals[:, None])
    else:
        F_mod = F_mod.at[dofs].set(vals)
    return K_mod, F_mod


def enforce_dirichlet_system(A, F, dofs, vals):
    """
    Apply Dirichlet conditions directly to stiffness/load.
    Dispatches based on matrix type (FluxSparseMatrix, JAX dense, or numpy dense).
    """
    if isinstance(A, FluxSparseMatrix):
        return enforce_dirichlet_sparse(A, F, dofs, vals)
    try:
        import jax.numpy as jnp

        if isinstance(A, jnp.ndarray):
            return enforce_dirichlet_dense_jax(A, F, dofs, vals)
    except Exception:
        pass
    return enforce_dirichlet_dense(A, F, dofs, vals)


def split_dirichlet_matrix(A, dir_dofs, *, n_total: int | None = None):
    """
    Split a matrix into free-free and free-dirichlet blocks.

    Returns (free, dir_dofs, A_ff, A_fd).
    """
    dir_dofs, _ = _normalize_dirichlet(dir_dofs, None)
    if n_total is None:
        if hasattr(A, "n_dofs"):
            n_total = int(A.n_dofs)
        else:
            arr = np.asarray(A)
            if arr.ndim != 2 or arr.shape[0] != arr.shape[1]:
                raise ValueError("A must be square when n_total is not provided.")
            n_total = int(arr.shape[0])
    free = free_dofs(n_total, dir_dofs)

    if isinstance(A, FluxSparseMatrix):
        if sp is None:
            raise ImportError("scipy is required to split FluxSparseMatrix.")
        A = A.to_csr()

    if sp is not None and sp.issparse(A):
        return free, dir_dofs, A[free][:, free], A[free][:, dir_dofs]

    if isinstance(A, jnp.ndarray):
        free_j = jnp.asarray(free, dtype=jnp.int32)
        dir_j = jnp.asarray(dir_dofs, dtype=jnp.int32)
        return free, dir_dofs, A[jnp.ix_(free_j, free_j)], A[jnp.ix_(free_j, dir_j)]

    arr = np.asarray(A)
    return free, dir_dofs, arr[np.ix_(free, free)], arr[np.ix_(free, dir_dofs)]


def condense_dirichlet_system(A, F, dofs, vals, *, check: bool = True) -> CondensedSystem:
    """
    Condense Dirichlet DOFs and return a structured system.
    """
    dir_arr, dir_vals_arr = _normalize_dirichlet(dofs, vals)
    F_arr = np.asarray(F)
    if hasattr(A, "n_dofs"):
        n_total = int(A.n_dofs)
    else:
        A_np = np.asarray(A)
        if A_np.ndim != 2 or A_np.shape[0] != A_np.shape[1]:
            raise ValueError("A must be square for Dirichlet condensation.")
        n_total = int(A_np.shape[0])

    if check:
        if dir_arr.size != dir_vals_arr.size:
            raise ValueError("dir_dofs and dir_vals must have the same length")
        if dir_arr.size:
            if np.min(dir_arr) < 0 or np.max(dir_arr) >= n_total:
                raise ValueError("dir_dofs out of bounds")
            if np.unique(dir_arr).size != dir_arr.size:
                raise ValueError("dir_dofs contains duplicates")

    mask = np.ones(n_total, dtype=bool)
    mask[dir_arr] = False
    free = np.nonzero(mask)[0]

    if isinstance(A, FluxSparseMatrix):
        K_csr = A.to_csr()
    elif sp is not None and sp.issparse(A):
        K_csr = A.tocsr()
    elif hasattr(A, "to_csr"):
        K_csr = A.to_csr()
    else:
        K_csr = np.asarray(A)

    K_ff = K_csr[free][:, free]
    F_free = F_arr[free]
    if dir_arr.size:
        K_fd = K_csr[free][:, dir_arr]
        if F_free.ndim == 2:
            F_free = F_free - (K_fd @ dir_vals_arr)[:, None]
        else:
            F_free = F_free - K_fd @ dir_vals_arr

    return CondensedSystem(
        K=K_ff,
        F=F_free,
        free_dofs=free,
        dir_dofs=dir_arr,
        dir_vals=dir_vals_arr,
        n_dofs=n_total,
    )




def enforce_dirichlet_sparse(A: FluxSparseMatrix, F, dofs, vals):
    """Apply Dirichlet conditions to FluxSparseMatrix + load (CSR)."""
    K_csr = A.to_csr().tolil()
    Fc = np.asarray(F, dtype=float).copy()
    dofs, vals = _normalize_dirichlet(dofs, vals)
    if Fc.ndim == 2:
        Fc = Fc - (K_csr[:, dofs] @ vals)[:, None]
    else:
        Fc = Fc - K_csr[:, dofs] @ vals
    for d, v in zip(dofs, vals):
        K_csr.rows[d] = [d]
        K_csr.data[d] = [1.0]
        K_csr[:, d] = 0.0
        K_csr[d, d] = 1.0
        if Fc.ndim == 2:
            Fc[d, :] = v
        else:
            Fc[d] = v
    return K_csr.tocsr(), Fc


def enforce_dirichlet_fluxsparse(A: FluxSparseMatrix, F, dofs, vals):
    """Alias for enforce_dirichlet_sparse for FluxSparseMatrix inputs."""
    return enforce_dirichlet_sparse(A, F, dofs, vals)


def condense_dirichlet_fluxsparse(A: FluxSparseMatrix, F, dofs, vals):
    """
    Condense Dirichlet DOFs for a FluxSparseMatrix.
    Returns: (K_ff, F_free, free_dofs, dir_dofs, dir_vals)
    """
    K_csr = A.to_csr()
    dir_arr, dir_vals_arr = _normalize_dirichlet(dofs, vals)
    mask = np.ones(K_csr.shape[0], dtype=bool)
    mask[dir_arr] = False
    free = np.nonzero(mask)[0]
    K_ff = K_csr[free][:, free]
    K_fd = K_csr[free][:, dir_arr] if dir_arr.size > 0 else None
    F_full = np.asarray(F, dtype=float)
    F_free = F_full[free]
    if K_fd is not None and dir_arr.size > 0:
        if F_free.ndim == 2:
            F_free = F_free - (K_fd @ dir_vals_arr)[:, None]
        else:
            F_free = F_free - K_fd @ dir_vals_arr
    return K_ff, F_free, free, dir_arr, dir_vals_arr


def condense_dirichlet_fluxsparse_coo(
    A: FluxSparseMatrix,
    F,
    dofs,
    vals,
    *,
    coalesce: bool = True,
):
    """
    Condense Dirichlet DOFs for a FluxSparseMatrix using COO filtering.
    Returns: (K_free, F_free, free_dofs, dir_dofs, dir_vals)
    """
    dir_arr, dir_vals_arr = _normalize_dirichlet(dofs, vals)
    n_total = int(A.n_dofs)
    mask = np.ones(n_total, dtype=bool)
    mask[dir_arr] = False
    free = np.nonzero(mask)[0]

    rows = np.asarray(A.pattern.rows, dtype=np.int64)
    cols = np.asarray(A.pattern.cols, dtype=np.int64)
    data = np.asarray(A.data)

    g2l = -np.ones(n_total, dtype=np.int64)
    g2l[free] = np.arange(free.size, dtype=np.int64)
    r2 = g2l[rows]
    c2 = g2l[cols]
    keep = (r2 >= 0) & (c2 >= 0)

    rows_f = r2[keep]
    cols_f = c2[keep]
    data_f = data[keep]
    if coalesce:
        rows_f, cols_f, data_f = coalesce_coo(rows_f, cols_f, data_f)

    K_free = FluxSparseMatrix(rows_f, cols_f, data_f, int(free.size))

    F_arr = np.asarray(F, dtype=float)
    F_free = F_arr[free]
    if dir_arr.size > 0 and not np.allclose(dir_vals_arr, 0.0):
        dir_full = np.zeros(n_total, dtype=F_arr.dtype)
        dir_full[dir_arr] = dir_vals_arr
        mask_fd = mask[rows] & (~mask[cols])
        if np.any(mask_fd):
            rows_fd = rows[mask_fd]
            cols_fd = cols[mask_fd]
            data_fd = data[mask_fd]
            contrib = data_fd * dir_full[cols_fd]
            delta = np.zeros(n_total, dtype=F_arr.dtype)
            np.add.at(delta, rows_fd, contrib)
            if F_free.ndim == 2:
                F_free = F_free - delta[free][:, None]
            else:
                F_free = F_free - delta[free]

    return K_free, jnp.asarray(F_free), free, dir_arr, dir_vals_arr


def free_dofs(n_dofs: int, dir_dofs) -> np.ndarray:
    """
    Return free DOF indices given total DOFs and Dirichlet DOFs.
    """
    dir_set = np.asarray(dir_dofs, dtype=int)
    mask = np.ones(int(n_dofs), dtype=bool)
    mask[dir_set] = False
    return np.nonzero(mask)[0]


def restrict_flux_to_free(K: FluxSparseMatrix, free: np.ndarray, *, coalesce: bool = True) -> FluxSparseMatrix:
    """
    Restrict a FluxSparseMatrix to free DOFs and return the condensed matrix.
    """
    free = np.asarray(free, dtype=np.int64)
    g2l = -np.ones(K.n_dofs, dtype=np.int64)
    g2l[free] = np.arange(free.size, dtype=np.int64)

    rows = np.asarray(K.pattern.rows)
    cols = np.asarray(K.pattern.cols)
    data = np.asarray(K.data)
    r2 = g2l[rows]
    c2 = g2l[cols]
    mask = (r2 >= 0) & (c2 >= 0)
    K_free = FluxSparseMatrix(
        jnp.asarray(r2[mask]),
        jnp.asarray(c2[mask]),
        jnp.asarray(data[mask]),
        int(free.size),
    )
    return K_free.coalesce() if coalesce else K_free


def condense_dirichlet_dense(K, F, dofs, vals):
    """
    Eliminate Dirichlet dofs for dense/CSR matrices and return condensed system.
    Returns: (K_cc, F_c, free_dofs, dir_dofs, dir_vals)
    """
    K_np = np.asarray(K, dtype=float)
    F_np = np.asarray(F, dtype=float)
    n = K_np.shape[0]

    dir_set, dir_vals = _normalize_dirichlet(dofs, vals)
    mask = np.ones(n, dtype=bool)
    mask[dir_set] = False
    free_dofs = np.nonzero(mask)[0]

    K_ff = K_np[np.ix_(free_dofs, free_dofs)]
    K_fd = K_np[np.ix_(free_dofs, dir_set)]
    F_f = F_np[free_dofs]
    if F_f.ndim == 2:
        F_f = F_f - (K_fd @ dir_vals)[:, None]
    else:
        F_f = F_f - K_fd @ dir_vals

    return K_ff, F_f, free_dofs, dir_set, dir_vals


def expand_dirichlet_solution(u_free, free_dofs, dir_dofs, dir_vals, n_total):
    """Expand condensed solution back to full vector."""
    dir_dofs, dir_vals = _normalize_dirichlet(dir_dofs, dir_vals)
    u_free_arr = np.asarray(u_free, dtype=float)
    if u_free_arr.ndim == 2:
        u = np.zeros((n_total, u_free_arr.shape[1]), dtype=float)
        u[free_dofs, :] = u_free_arr
        u[dir_dofs, :] = np.asarray(dir_vals, dtype=float)
    else:
        u = np.zeros(n_total, dtype=float)
        u[free_dofs] = u_free_arr
        u[dir_dofs] = np.asarray(dir_vals, dtype=float)
    return u



from __future__ import annotations

from typing import Callable, Iterable, Sequence, TypeAlias

import numpy as np
import jax.numpy as jnp

try:
    from jax.experimental import sparse as jsparse
except Exception:  # pragma: no cover
    jsparse = None

from .sparse import FluxSparseMatrix

ArrayLike: TypeAlias = jnp.ndarray
Preconditioner: TypeAlias = Callable[[jnp.ndarray], jnp.ndarray]


def _extract_block_sizes(
    n: int,
    *,
    dof_per_node: int | None,
    block_sizes: Sequence[int] | None,
    meta: dict | None,
) -> int:
    sizes = None
    if block_sizes is not None:
        sizes = np.asarray(block_sizes, dtype=int)
    elif dof_per_node is not None:
        if dof_per_node <= 0:
            raise ValueError("dof_per_node must be positive")
        if n % dof_per_node != 0:
            raise ValueError("dof_per_node must divide n_dofs")
        sizes = np.full(n // dof_per_node, dof_per_node, dtype=int)
    elif meta:
        if meta.get("dof_layout") not in (None, "blocked"):
            raise ValueError("block_jacobi requires dof_layout='blocked'")
        if "block_sizes" in meta:
            sizes = np.asarray(meta["block_sizes"], dtype=int)
        elif "dof_per_node" in meta:
            d = int(meta["dof_per_node"])
            if d <= 0 or n % d != 0:
                raise ValueError("meta['dof_per_node'] must divide n_dofs")
            sizes = np.full(n // d, d, dtype=int)
    if sizes is None or sizes.size == 0:
        raise ValueError("block_jacobi requires block_sizes or dof_per_node")
    if np.any(sizes <= 0):
        raise ValueError("block_sizes entries must be positive")
    if int(sizes.sum()) != n:
        raise ValueError("sum(block_sizes) must equal n_dofs")
    if not np.all(sizes == sizes[0]):
        raise ValueError("block_sizes must be uniform for block_jacobi")
    return int(sizes[0])


def make_block_jacobi_preconditioner(
    A: FluxSparseMatrix | "jsparse.BCOO",
    *,
    dof_per_node: int | None = None,
    block_sizes: Sequence[int] | None = None,
) -> Preconditioner:
    """
    Build block Jacobi preconditioner for blocked DOF layouts.

    Priority:
      block_sizes -> dof_per_node -> A.meta (block_sizes / dof_per_node) -> error
    """
    if jsparse is not None and isinstance(A, jsparse.BCOO):
        n = int(A.shape[0])
        block_size = _extract_block_sizes(
            n, dof_per_node=dof_per_node, block_sizes=block_sizes, meta=None
        )
        rows = jnp.asarray(A.indices[:, 0])
        cols = jnp.asarray(A.indices[:, 1])
        data = jnp.asarray(A.data)
    elif isinstance(A, FluxSparseMatrix):
        n = int(A.n_dofs)
        block_size = _extract_block_sizes(
            n, dof_per_node=dof_per_node, block_sizes=block_sizes, meta=A.meta
        )
        rows = jnp.asarray(A.pattern.rows)
        cols = jnp.asarray(A.pattern.cols)
        data = jnp.asarray(A.data)
    else:
        raise ValueError("block_jacobi requires FluxSparseMatrix or BCOO")

    if n % block_size != 0:
        raise ValueError("block_size must divide n_dofs")
    n_block = n // block_size
    block_rows = rows // block_size
    block_cols = cols // block_size
    lr = rows % block_size
    lc = cols % block_size
    mask = block_rows == block_cols
    block_rows = block_rows[mask]
    lr = lr[mask]
    lc = lc[mask]
    data = data[mask]
    blocks = jnp.zeros((n_block, block_size, block_size), dtype=data.dtype)
    blocks = blocks.at[block_rows, lr, lc].add(data)
    blocks = blocks + 1e-12 * jnp.eye(block_size)[None, :, :]
    inv_blocks = jnp.linalg.inv(blocks)

    def precon(r: jnp.ndarray) -> jnp.ndarray:
        rb = r.reshape((n_block, block_size))
        zb = jnp.einsum("bij,bj->bi", inv_blocks, rb)
        return zb.reshape((-1,))

    return precon



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Iterable, Sequence, TYPE_CHECKING, TypeAlias

import numpy as np
import jax
import jax.numpy as jnp

try:
    import scipy.sparse as sp
except Exception:  # pragma: no cover
    sp = None


if TYPE_CHECKING:
    from jax import Array as JaxArray

    ArrayLike: TypeAlias = np.ndarray | JaxArray
else:
    ArrayLike: TypeAlias = np.ndarray
COOTuple: TypeAlias = tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, int]
INDEX_DTYPE = jnp.int64 if jax.config.read("jax_enable_x64") else jnp.int32


def coalesce_coo(
    rows: ArrayLike, cols: ArrayLike, data: ArrayLike
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Sum duplicate COO entries by sorting (CPU-friendly).
    Returns (rows_u, cols_u, data_u) as NumPy arrays.
    """
    r = np.asarray(rows, dtype=np.int64)
    c = np.asarray(cols, dtype=np.int64)
    d = np.asarray(data)
    if r.size == 0:
        return r, c, d
    order = np.lexsort((c, r))
    r_s = r[order]
    c_s = c[order]
    d_s = d[order]
    new_group = np.ones(r_s.size, dtype=bool)
    new_group[1:] = (r_s[1:] != r_s[:-1]) | (c_s[1:] != c_s[:-1])
    starts = np.nonzero(new_group)[0]
    r_u = r_s[starts]
    c_u = c_s[starts]
    d_u = np.add.reduceat(d_s, starts)
    return r_u, c_u, d_u


def _normalize_flux_mats(mats: Sequence["FluxSparseMatrix"]) -> tuple["FluxSparseMatrix", ...]:
    if len(mats) == 1 and isinstance(mats[0], (list, tuple)):
        mats = tuple(mats[0])
    if not mats:
        raise ValueError("At least one FluxSparseMatrix is required.")
    return mats


def concat_flux(*mats: "FluxSparseMatrix", n_dofs: int | None = None) -> "FluxSparseMatrix":
    """
    Concatenate COO entries from multiple FluxSparseMatrix objects.
    All matrices must share the same n_dofs unless n_dofs is provided.
    """
    mats = _normalize_flux_mats(mats)
    if n_dofs is None:
        n_dofs = int(mats[0].n_dofs)
        for mat in mats[1:]:
            if int(mat.n_dofs) != n_dofs:
                raise ValueError("All matrices must share n_dofs for concat_flux.")
    rows_list = [np.asarray(mat.pattern.rows, dtype=np.int64) for mat in mats]
    cols_list = [np.asarray(mat.pattern.cols, dtype=np.int64) for mat in mats]
    data_list = [np.asarray(mat.data) for mat in mats]
    rows = np.concatenate(rows_list) if rows_list else np.asarray([], dtype=np.int64)
    cols = np.concatenate(cols_list) if cols_list else np.asarray([], dtype=np.int64)
    data = np.concatenate(data_list) if data_list else np.asarray([], dtype=float)
    return FluxSparseMatrix(rows, cols, data, int(n_dofs))


def block_diag_flux(*mats: "FluxSparseMatrix") -> "FluxSparseMatrix":
    """Block-diagonal concatenation for FluxSparseMatrix objects."""
    mats = _normalize_flux_mats(mats)
    rows_out = []
    cols_out = []
    data_out = []
    offset = 0
    for mat in mats:
        rows = np.asarray(mat.pattern.rows, dtype=np.int64)
        cols = np.asarray(mat.pattern.cols, dtype=np.int64)
        data = np.asarray(mat.data)
        if rows.size:
            rows_out.append(rows + offset)
            cols_out.append(cols + offset)
            data_out.append(data)
        offset += int(mat.n_dofs)
    rows = np.concatenate(rows_out) if rows_out else np.asarray([], dtype=np.int64)
    cols = np.concatenate(cols_out) if cols_out else np.asarray([], dtype=np.int64)
    data = np.concatenate(data_out) if data_out else np.asarray([], dtype=float)
    return FluxSparseMatrix(rows, cols, data, int(offset))


@jax.tree_util.register_pytree_node_class
@dataclass(frozen=True)
class SparsityPattern:
    """
    Jacobian sparsity pattern (rows/cols) that is independent of the solution.
    """

    rows: jnp.ndarray
    cols: jnp.ndarray
    n_dofs: int
    idx: jnp.ndarray | None = None
    diag_idx: jnp.ndarray | None = None
    perm: jnp.ndarray | None = None        # permutation mapping COO data -> CSR data
    indptr: jnp.ndarray | None = None      # CSR row pointer
    indices: jnp.ndarray | None = None     # CSR column indices

    def __post_init__(self):
        # Ensure n_dofs is always a Python int so JAX treats it as a static aux value.
        object.__setattr__(self, "n_dofs", int(self.n_dofs))

    def tree_flatten(self):
        children = (
            self.rows,
            self.cols,
            self.idx if self.idx is not None else jnp.array([], INDEX_DTYPE),
            self.diag_idx if self.diag_idx is not None else jnp.array([], INDEX_DTYPE),
            self.perm if self.perm is not None else jnp.array([], INDEX_DTYPE),
            self.indptr if self.indptr is not None else jnp.array([], INDEX_DTYPE),
            self.indices if self.indices is not None else jnp.array([], INDEX_DTYPE),
        )
        aux = {
            "n_dofs": self.n_dofs,
            "has_idx": self.idx is not None,
            "has_diag_idx": self.diag_idx is not None,
            "has_perm": self.perm is not None,
            "has_indptr": self.indptr is not None,
            "has_indices": self.indices is not None,
        }
        return children, aux

    @classmethod
    def tree_unflatten(cls, aux, children):
        rows, cols, idx, diag_idx, perm, indptr, indices = children
        idx_out = idx if aux["has_idx"] else None
        diag_out = diag_idx if aux["has_diag_idx"] else None
        perm_out = perm if aux["has_perm"] else None
        indptr_out = indptr if aux["has_indptr"] else None
        indices_out = indices if aux["has_indices"] else None
        return cls(
            rows=rows,
            cols=cols,
            n_dofs=aux["n_dofs"],
            idx=idx_out,
            diag_idx=diag_out,
            perm=perm_out,
            indptr=indptr_out,
            indices=indices_out,
        )


@jax.tree_util.register_pytree_node_class
class FluxSparseMatrix:
    """
    Sparse matrix wrapper (COO) with a fixed pattern and mutable values.
    - pattern stores rows/cols/n_dofs (optionally idx for dense scatter)
    - data stores the numeric values for the current nonlinear iterate
    """

    def __init__(
        self,
        rows_or_pattern: SparsityPattern | ArrayLike,
        cols: ArrayLike | None = None,
        data: ArrayLike | None = None,
        n_dofs: int | None = None,
        meta: dict | None = None,
    ):
        # New signature: FluxSparseMatrix(pattern, data)
        if isinstance(rows_or_pattern, SparsityPattern):
            pattern = rows_or_pattern
            values = cols if data is None else data
            values = jnp.asarray(values)
        else:
            # Legacy signature: FluxSparseMatrix(rows, cols, data, n_dofs)
            r_j = jnp.asarray(rows_or_pattern, dtype=INDEX_DTYPE)
            c_j = jnp.asarray(cols, dtype=INDEX_DTYPE)
            is_tracer = isinstance(rows_or_pattern, jax.core.Tracer) or isinstance(cols, jax.core.Tracer)
            diag_idx_j = None
            if not is_tracer:
                diag_idx_j = jnp.nonzero(r_j == c_j)[0].astype(INDEX_DTYPE)
            if n_dofs is None:
                if is_tracer:
                    raise ValueError("n_dofs must be provided when constructing FluxSparseMatrix under JIT.")
                n_dofs = int(np.asarray(cols).max()) + 1
            pattern = SparsityPattern(
                rows=r_j,
                cols=c_j,
                n_dofs=int(n_dofs) if n_dofs is not None else int(np.asarray(cols).max()) + 1,
                idx=None,
                diag_idx=diag_idx_j,
            )
            values = jnp.asarray(data)

        self.pattern = pattern
        self.rows = pattern.rows
        self.cols = pattern.cols
        self.n_dofs = int(pattern.n_dofs)
        self.data = values
        self.meta = dict(meta) if meta is not None else None

    @classmethod
    def from_bilinear(cls, coo_tuple: COOTuple) -> "FluxSparseMatrix":
        """Construct from assemble_bilinear_dense(..., sparse=True)."""
        rows, cols, data, n_dofs = coo_tuple
        return cls(rows, cols, data, n_dofs)

    @classmethod
    def from_linear(cls, coo_tuple: tuple[jnp.ndarray, jnp.ndarray, int]) -> "FluxSparseMatrix":
        """Construct from assemble_linear_form(..., sparse=True) (matrix interpretation only)."""
        rows, data, n_dofs = coo_tuple
        cols = jnp.zeros_like(rows)
        return cls(rows, cols, data, n_dofs)

    def with_data(self, data: ArrayLike) -> "FluxSparseMatrix":
        """Return a new FluxSparseMatrix sharing the same pattern with updated data."""
        return FluxSparseMatrix(self.pattern, data, meta=self.meta)

    def add_dense(self, dense: ArrayLike) -> "FluxSparseMatrix":
        """Return a new FluxSparseMatrix with dense entries added on the pattern."""
        dense_vals = jnp.asarray(dense)[self.pattern.rows, self.pattern.cols]
        return FluxSparseMatrix(self.pattern, self.data + dense_vals)

    def to_coo(self) -> COOTuple:
        return self.pattern.rows, self.pattern.cols, self.data, self.pattern.n_dofs

    @property
    def nnz(self) -> int:
        return int(self.data.shape[0])

    def coalesce(self) -> "FluxSparseMatrix":
        """Return a new FluxSparseMatrix with duplicate entries summed."""
        rows_u, cols_u, data_u = coalesce_coo(self.pattern.rows, self.pattern.cols, self.data)
        return FluxSparseMatrix(rows_u, cols_u, data_u, self.pattern.n_dofs)

    def to_csr(self):
        if sp is None:
            raise ImportError("scipy is required for to_csr()")
        if (
            self.pattern.indptr is not None
            and self.pattern.indices is not None
            and self.pattern.perm is not None
        ):
            indptr = np.array(self.pattern.indptr, dtype=np.int64, copy=True)
            indices = np.array(self.pattern.indices, dtype=np.int64, copy=True)
            data = np.array(self.data, copy=True)[np.asarray(self.pattern.perm, dtype=np.int64)]
            return sp.csr_matrix((data, indices, indptr), shape=(self.pattern.n_dofs, self.pattern.n_dofs))
        r = np.array(self.pattern.rows, dtype=np.int64, copy=True)
        c = np.array(self.pattern.cols, dtype=np.int64, copy=True)
        d = np.array(self.data, copy=True)
        return sp.csr_matrix((d, (r, c)), shape=(self.pattern.n_dofs, self.pattern.n_dofs))

    def to_dense(self) -> jnp.ndarray:
        # small debug helper
        dense = jnp.zeros((self.pattern.n_dofs, self.pattern.n_dofs), dtype=self.data.dtype)
        dense = dense.at[self.pattern.rows, self.pattern.cols].add(self.data)
        return dense

    def to_bcoo(self):
        """Construct jax.experimental.sparse.BCOO (requires jax.experimental.sparse)."""
        try:
            from jax.experimental import sparse as jsparse  # type: ignore
        except Exception as exc:  # pragma: no cover
            raise ImportError("jax.experimental.sparse is required for to_bcoo()") from exc
        idx = jnp.stack([self.pattern.rows, self.pattern.cols], axis=-1)
        return jsparse.BCOO((self.data, idx), shape=(self.pattern.n_dofs, self.pattern.n_dofs))

    def matvec(self, x: ArrayLike) -> jnp.ndarray:
        """Compute y = A x in JAX (iterative solvers)."""
        xj = jnp.asarray(x)
        contrib = self.data * xj[self.pattern.cols]
        # Use scatter_add to avoid tracing a dynamic int(x.max()) in jnp.bincount,
        # which triggers concretization errors under jit/while_loop.
        out = jnp.zeros(self.pattern.n_dofs, dtype=contrib.dtype)
        return out.at[self.pattern.rows].add(contrib)

    def as_cg_operator(
        self,
        *,
        matvec: str = "flux",
        preconditioner=None,
        solver: str = "cg",
        dof_per_node: int | None = None,
        block_sizes=None,
    ):
        from .cg import build_cg_operator

        return build_cg_operator(
            self,
            matvec=matvec,
            preconditioner=preconditioner,
            solver=solver,
            dof_per_node=dof_per_node,
            block_sizes=block_sizes,
        )

    def diag(self):
        """Diagonal entries aggregated for Jacobi preconditioning."""
        if self.pattern.diag_idx is not None:
            r = self.pattern.rows[self.pattern.diag_idx]
            d = self.data[self.pattern.diag_idx]
            return jax.ops.segment_sum(d, r, self.pattern.n_dofs)

        # Fallback for patterns without diag_idx (kept for backward compatibility).
        mask = self.pattern.rows == self.pattern.cols
        diag_contrib = jnp.where(mask, self.data, jnp.zeros_like(self.data))
        return jax.ops.segment_sum(diag_contrib, self.pattern.rows, self.pattern.n_dofs)

    def tree_flatten(self):
        return (self.pattern, self.data), {}

    @classmethod
    def tree_unflatten(cls, aux, children):
        pattern, data = children
        return cls(pattern, data)



from __future__ import annotations

import importlib

__all__ = [
    "FESpaceBase",
    "FESpace",
    "FESpacePytree",
    "Expr",
    "FieldRef",
    "ParamRef",
    "trial_ref",
    "test_ref",
    "unknown_ref",
    "zero_ref",
    "Params",
    "LinearForm",
    "BilinearForm",
    "ResidualForm",
    "MixedWeakForm",
    "make_mixed_residuals",
    "kernel",
    "MixedFESpace",
    "MixedProblem",
    "MixedDirichletBC",
    "MixedBlockSystem",
    "compile_bilinear",
    "compile_linear",
    "compile_residual",
    "compile_surface_linear",
    "compile_surface_bilinear",
    "compile_mixed_surface_residual",
    "compile_mixed_residual",
    "FormContext",
    "MixedFormContext",
    "VolumeContext",
    "SurfaceContext",
    "FieldPair",
    "ElementVector",
    "vector_load_form",
    "make_space",
    "make_space_pytree",
    "make_hex_basis",
    "make_hex_basis_pytree",
    "make_hex_space",
    "make_hex_space_pytree",
    "make_hex20_basis",
    "make_hex20_basis_pytree",
    "make_hex20_space",
    "make_hex20_space_pytree",
    "make_hex27_basis",
    "make_hex27_basis_pytree",
    "make_hex27_space",
    "make_hex27_space_pytree",
    "make_tet_basis",
    "make_tet_basis_pytree",
    "make_tet_space",
    "make_tet_space_pytree",
    "make_tet10_basis",
    "make_tet10_basis_pytree",
    "make_tet10_space",
    "make_tet10_space_pytree",
    "MeshData",
    "BasisData",
    "SpaceData",
    "make_element_residual_kernel",
    "make_element_bilinear_kernel",
    "make_element_linear_kernel",
    "make_element_jacobian_kernel",
    "make_element_kernel",
    "AssemblyPolicy",
    "element_residual",
    "element_jacobian",
    "make_sparsity_pattern",
    "chunk_pad_stats",
    "BatchedAssembler",
    "assemble_functional",
    "assemble_mass_matrix",
    "scalar_body_force_form",
    "make_scalar_body_force_form",
    "vector_body_force_form",
    "linear_elasticity_form",
    "constant_body_force_form",
    "constant_body_force_vector_form",
    "diffusion_form",
    "right_cauchy_green",
    "green_lagrange_strain",
    "deformation_gradient",
    "pk2_neo_hookean",
    "neo_hookean_residual_form",
    "StokesSpaces",
    "make_stokes_spaces",
    "HexMesh",
    "HexMeshPytree",
    "StructuredHexBox",
    "SurfaceMesh",
    "SurfaceMeshPytree",
    "SurfaceWithElemConn",
    "surface_with_elem_conn",
    "SurfaceFormField",
    "SurfaceFormContext",
    "SurfaceBilinearContext",
    "SurfaceSupermesh",
    "vector_surface_load_form",
    "make_vector_surface_load_form",
    "assemble_surface_linear_form",
    "assemble_surface_bilinear_form",
    "ContactCouplingMatrix",
    "assemble_contact_coupling_matrices",
    "assemble_contact_onesided_floor",
    "assemble_contact_interface_residual",
    "assemble_contact_interface_jacobian",
    "map_surface_facets_to_tet_elements",
    "map_surface_facets_to_hex_elements",
    "tri_area",
    "tri_quadrature",
    "facet_triangles",
    "facet_shape_values",
    "volume_shape_values_at_points",
    "quad_shape_and_local",
    "quad9_shape_values",
    "hex27_gradN",
    "ContactSurfaceSpace",
    "ContactSide",
    "ContactOperators",
    "OneSidedContact",
    "OneSidedContactSurfaceSpace",
    "OneToManyContactSurfaceSpace",
    "assemble_contact_operators",
    "assemble_contact_interface_residual",
    "assemble_contact_interface_jacobian",
    "assemble_contact_coupling_matrices",
    "assemble_contact_kkt",
    "solve_contact_kkt",
    "facet_gap_values",
    "active_contact_facets",
    "tag_axis_minmax_facets",
    "load_gmsh_mesh",
    "load_gmsh_hex_mesh",
    "load_gmsh_tet_mesh",
    "make_surface_from_facets",
    "build_surface_supermesh",
    "ContactCouplingMatrix",
    "assemble_contact_coupling_matrices",
    "assemble_contact_interface_residual",
    "assemble_contact_interface_jacobian",
    "map_surface_facets_to_tet_elements",
    "TetMesh",
    "TetMeshPytree",
    "StructuredTetBox",
    "StructuredTetTensorBox",
    "BaseMeshPytree",
    "SurfaceWithFacetMap",
    "bbox_predicate",
    "plane_predicate",
    "axis_plane_predicate",
    "slab_predicate",
    "HexTriLinearBasis",
    "HexTriLinearBasisPytree",
    "HexSerendipityBasis20",
    "HexSerendipityBasis20Pytree",
    "HexTriQuadraticBasis27",
    "HexTriQuadraticBasis27Pytree",
    "TetLinearBasis",
    "TetLinearBasisPytree",
    "TetQuadraticBasis10",
    "TetQuadraticBasis10Pytree",
    "lame_parameters",
    "isotropic_3d_D",
    "spdirect_solve_cpu",
    "spdirect_solve_gpu",
    "spdirect_solve_jax",
    "coo_to_csr",
    "SparsityPattern",
    "FluxSparseMatrix",
    "FluxBlockMatrix",
    "coalesce_coo",
    "concat_flux",
    "block_diag_flux",
    "build_block_system",
    "split_block_matrix",
    "BlockSystem",
    "LinearSolver",
    "NonlinearSolver",
    "NewmarkResult",
    "newmark_solve_linear",
    "petsc_solve",
    "petsc_shell_solve",
    "petsc_is_available",
    "DirichletBC",
    "enforce_dirichlet_dense",
    "enforce_dirichlet_dense_jax",
    "enforce_dirichlet_fluxsparse",
    "enforce_dirichlet_sparse",
    "enforce_dirichlet_system",
    "split_dirichlet_matrix",
    "free_dofs",
    "condense_dirichlet_fluxsparse",
    "condense_dirichlet_fluxsparse_coo",
    "condense_dirichlet_dense",
    "restrict_flux_to_free",
    "expand_dirichlet_solution",
    "condense_dirichlet_system",
    "cg_solve",
    "make_block_jacobi_preconditioner",
    "build_cg_operator",
    "CGOperator",
    "NonlinearAnalysis",
    "NewtonLoopConfig",
    "LoadStepResult",
    "NewtonSolveRunner",
    "solve_nonlinear",
    "LinearAnalysis",
    "LinearSolveConfig",
    "LinearStepResult",
    "LinearSolveRunner",
    "newton_solve",
    "write_vtu",
    "write_displacement_vtu",
    "make_jitted_residual",
    "make_jitted_jacobian",
    "block",
    "make_elastic_point_data",
    "write_elastic_vtu",
    "make_point_data_displacement",
    "write_point_data_vtu",
    "principal_stresses",
    "principal_sum",
    "max_shear_stress",
    "von_mises_stress",
]

_PHYSICS_EXPORTS = {
    "lame_parameters",
    "isotropic_3d_D",
    "linear_elasticity_form",
    "vector_body_force_form",
    "constant_body_force_vector_form",
    "diffusion_form",
    "right_cauchy_green",
    "green_lagrange_strain",
    "deformation_gradient",
    "pk2_neo_hookean",
    "neo_hookean_residual_form",
    "make_elastic_point_data",
    "write_elastic_vtu",
    "make_point_data_displacement",
    "write_point_data_vtu",
    "interpolate_at_points",
    "principal_stresses",
    "principal_sum",
    "max_shear_stress",
    "von_mises_stress",
}

_TOOLS_VIS_EXPORTS = {
    "write_vtu",
    "write_displacement_vtu",
}

_TOOLS_JIT_EXPORTS = {
    "make_jitted_residual",
    "make_jitted_jacobian",
}

_CORE_EXPORTS = set(__all__) - _PHYSICS_EXPORTS - _TOOLS_VIS_EXPORTS - _TOOLS_JIT_EXPORTS


def __getattr__(name: str):
    if name in _PHYSICS_EXPORTS:
        module = importlib.import_module(".physics", __name__)
    elif name in _TOOLS_VIS_EXPORTS:
        module = importlib.import_module(".tools.visualizer", __name__)
    elif name in _TOOLS_JIT_EXPORTS:
        module = importlib.import_module(".tools.jit", __name__)
    elif name in _CORE_EXPORTS:
        module = importlib.import_module(".core", __name__)
    else:
        raise AttributeError(f"module {__name__!r} has no attribute {name!r}")

    value = getattr(module, name)
    globals()[name] = value
    return value


def __dir__():
    return sorted(set(globals()) | set(__all__))


try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    from importlib_metadata import version, PackageNotFoundError


def read_version_from_pyproject():
    import pathlib
    import re

    root = pathlib.Path(__file__).resolve().parent.parent
    pyproject_path = root / "pyproject.toml"
    if pyproject_path.exists():
        content = pyproject_path.read_text()
        match = re.search(r'version\s*=\s*"(.*?)"', content)
        if match:
            return match.group(1)
    return "0.0.0"


def get_version(package_name):
    try:
        return version(package_name)
    except PackageNotFoundError:
        return read_version_from_pyproject()


__version__ = get_version("fluxfem")
